import time
import os
import numpy as np
import sys
import click
from cosmoglobe.release.tools import *
from cosmoglobe.release.commands_plotting import *
from cosmoglobe.release.fitsformatter import format_fits, get_data, get_header


@click.group()
def commands_hdf():
    pass


@commands_hdf.command()
@click.argument("input", type=click.STRING)
@click.argument("dataset", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.option(
    "-min",
    default=1,
    type=click.INT,
    help="Start sample, default 1",
)
@click.option(
    "-max",
    default=None,
    type=click.INT,
    help="End sample, calculated automatically if not set",
)
@click.option(
    "-maxchain",
    default=1,
    help="max number of chains c0005 [ex. 5]",
)
@click.option(
    "-notchain",
    is_flag=True,
    help="Flag if parsing a non-chain hdf file",
)
def split(input, dataset, output, min, max, maxchain, notchain):
    """
    This function saves whatever specified data to a separate file.
    """
    h5handler(
        input,
        dataset,
        min,
        max,
        maxchain,
        output,
        fwhm=None,
        nside=None,
        command=False,
        pixweight=None,
        zerospin=False,
        lowmem=False,
        notchain=True,
    )


@commands_hdf.command()
@click.argument("input", type=click.STRING)
@click.argument("dataset", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.option(
    "-min",
    default=1,
    type=click.INT,
    help="Start sample, default 1",
)
@click.option(
    "-max",
    default=None,
    type=click.INT,
    help="End sample, calculated automatically if not set",
)
@click.option(
    "-maxchain",
    default=1,
    help="max number of chains c0005 [ex. 5]",
)
@click.option("-fwhm", default=0.0, help="FWHM in arcmin")
@click.option(
    "-nside",
    default=None,
    type=click.INT,
    help="Nside for alm binning",
)
@click.option(
    "-zerospin",
    is_flag=True,
    help="If smoothing, treat maps as zero-spin maps.",
)
@click.option(
    "-pixweight",
    default=None,
    type=click.STRING,
    help="Path to healpy pixel weights.",
)
def mean(input, dataset, output, min, max, maxchain, fwhm, nside, zerospin, pixweight):
    """
    Calculates the mean over sample range from .h5 file.
    ex. chains_c0001.h5 dust/amp_map 5 50 dust_5-50_mean_40arcmin.fits -fwhm 40 -maxchain 3
    ex. chains_c0001.h5 dust/amp_alm 5 50 dust_5-50_mean_40arcmin.fits -fwhm 40 -nside 512
    If output name is set to .dat, data will not be converted to map.
    """
    if dataset.endswith("alm") and nside == None:
        click.echo("Please specify nside when handling alms.")
        sys.exit()

    h5handler(
        input,
        dataset,
        min,
        max,
        maxchain,
        output,
        fwhm,
        nside,
        np.mean,
        pixweight,
        zerospin,
    )


@commands_hdf.command()
@click.argument("input", type=click.STRING)
@click.argument("dataset", type=click.STRING)
@click.argument("output", type=click.STRING)
@click.option(
    "-min",
    default=1,
    type=click.INT,
    help="Start sample, default 1",
)
@click.option(
    "-max",
    default=None,
    type=click.INT,
    help="End sample, calculated automatically if not set",
)
@click.option(
    "-maxchain",
    default=1,
    help="max number of chains c0005 [ex. 5]",
)
@click.option("-fwhm", default=0.0, help="FWHM in arcmin")
@click.option(
    "-nside",
    default=None,
    type=click.INT,
    help="Nside for alm binning",
)
@click.option(
    "-zerospin",
    is_flag=True,
    help="If smoothing, treat maps as zero-spin maps.",
)
@click.option(
    "-pixweight",
    default=None,
    type=click.STRING,
    help="Path to healpy pixel weights.",
)
def stddev(
    input,
    dataset,
    output,
    min,
    max,
    maxchain,
    fwhm,
    nside,
    zerospin,
    pixweight,
):
    """
    Calculates the stddev over sample range from .h5 file.
    ex. chains_c0001.h5 dust/amp_map 5 50 dust_5-50_mean_40arcmin.fits -fwhm 40 -maxchain 3
    ex. chains_c0001.h5 dust/amp_alm 5 50 dust_5-50_mean_40arcmin.fits -fwhm 40 -nside 512

    If output name is set to .dat, data will not be converted to map.
    """
    if dataset.endswith("alm") and nside == None:
        click.echo("Please specify nside when handling alms.")
        sys.exit()

    h5handler(
        input,
        dataset,
        min,
        max,
        maxchain,
        output,
        fwhm,
        nside,
        np.std,
        pixweight,
        zerospin,
    )


@commands_hdf.command()
@click.argument("filename", type=click.STRING)
@click.argument("nchains", type=click.INT)
@click.argument("burnin", type=click.INT)
@click.option(
    "-suffix", type=click.STRING, default="", help="if there is some, ex. Tresamp_v1"
)
@click.option(
    "-path",
    default="cmb/sigma_l",
    help="Dataset path ex. cmb/sigma_l",
)
@click.argument("outname", type=click.STRING)
def sigma_l2fits(
    filename,
    nchains,
    burnin,
    path,
    suffix,
    outname,
    save=True,
):
    """
    Converts c3-h5 dataset to fits for c1 BR and GBR estimator analysis.
    ex. c3pp sigma-l2fits chains_v1/chain 5 10 cmb_sigma_l_GBRlike.fits
    If "chain_c0001.h5", filename is cut to "chain" and will look in same directory for "chain_c*****.h5". If a -suffix is present, it is added after c****, ex. -suffix Tr\
esamp_v1 will point to chain_c****_Tresamp_v1.h5.
    See comm_like_tools for further information about BR and GBR post processing
    """
    click.echo("{:-^48}".format("Formatting sigma_l data to fits file"))
    import h5py

    if filename.endswith(".h5"):
        filename = filename.rsplit("_", 1)[0]
    temp = np.zeros(nchains)
    for nc in range(1, nchains + 1):
        with h5py.File(
            filename + "_c" + str(nc).zfill(4) + suffix + ".h5",
            "r",
        ) as f:
            groups = list(f.keys())
            temp[nc - 1] = len(groups)
    nsamples_max = int(max(temp[:]))
    click.echo(f"Largest chain has {nsamples_max} samples, using burnin {burnin}\n")
    for nc in range(1, nchains + 1):
        fn = filename + "_c" + str(nc).zfill(4) + suffix + ".h5"
        with h5py.File(
            fn,
            "r",
        ) as f:
            click.echo(f"Reading {fn}")
            groups = list(f.keys())
            nsamples = len(groups)
            if nc == 1:
                dset = np.zeros(
                    (
                        nsamples_max + 1,
                        1,
                        len(f[groups[0] + "/" + path]),
                        len(f[groups[0] + "/" + path][0]),
                    )
                )
                nspec = len(f[groups[0] + "/" + path])
                lmax = len(f[groups[0] + "/" + path][0]) - 1
            else:
                dset = np.append(
                    dset,
                    np.zeros(
                        (
                            nsamples_max + 1,
                            1,
                            nspec,
                            lmax + 1,
                        )
                    ),
                    axis=1,
                )
            click.echo(
                f"Dataset: {path} \n# samples: {nsamples} \n# spectra: {nspec} \nlmax: {lmax}"
            )
            for i in range(nsamples):
                for j in range(nspec):
                    dset[i + 1, nc - 1, j, :] = np.asarray(
                        f[groups[i] + "/" + path][j][:]
                    )
            click.echo("")
    # Optimize with jit?
    ell = np.arange(lmax + 1)
    for nc in range(1, nchains + 1):
        for i in range(1, nsamples_max + 1):
            for j in range(nspec):
                dset[i, nc - 1, j, :] = (
                    dset[i, nc - 1, j, :] * ell[:] * (ell[:] + 1.0) / 2.0 / np.pi
                )
    dset[0, :, :, :] = nsamples - 2  # burnin

    if save:
        click.echo(f"Dumping fits file: {outname}")
        dset = np.asarray(dset, dtype="f4")

        from astropy.io import fits

        head = fits.Header()
        head["FUNCNAME"] = ("Gibbs sampled power spectra", "Full function name")
        head["LMAX"] = (lmax, "Maximum multipole moment")
        head["NUMSAMP"] = (nsamples_max, "Number of samples")
        head["NUMCHAIN"] = (nchains, "Number of independent chains")
        head["NUMSPEC"] = (nspec, "Number of power spectra")
        fits.writeto(outname, dset, head, overwrite=True)

        # FITSIO Saving Deprecated (Use astropy)
        if False:
            import fitsio

            fits = fitsio.FITS(
                outname,
                mode="rw",
                clobber=True,
                verbose=True,
            )
            h_dict = [
                {
                    "name": "FUNCNAME",
                    "value": "Gibbs sampled power spectra",
                    "comment": "Full function name",
                },
                {
                    "name": "LMAX",
                    "value": lmax,
                    "comment": "Maximum multipole moment",
                },
                {
                    "name": "NUMSAMP",
                    "value": nsamples_max,
                    "comment": "Number of samples",
                },
                {
                    "name": "NUMCHAIN",
                    "value": nchains,
                    "comment": "Number of independent chains",
                },
                {
                    "name": "NUMSPEC",
                    "value": nspec,
                    "comment": "Number of power spectra",
                },
            ]
            fits.write(
                dset[:, :, :, :],
                header=h_dict,
                clobber=True,
            )
            fits.close()

    return dset


def h5map2fits(filename, dataset, save=True):
    """
    Outputs a .h5 map to fits on the form 000001_cmb_amp_n1024.fits
    """
    import healpy as hp
    import h5py

    dataset, tag = dataset.rsplit("/", 1)

    with h5py.File(filename, "r") as f:
        maps = f[f"{dataset}/{tag}"][()]
        lmax = f[f"{dataset}/amp_lmax"][()]  # Get lmax from h5

    nside = hp.npix2nside(maps.shape[-1])
    dataset = f"{dataset}/{tag}"
    outfile = dataset.replace("/", "_")
    outfile = outfile.replace("_map", "")
    if save:
        hp.write_map(outfile + f"_n{str(nside)}.fits", maps, overwrite=True, dtype=None)
    return maps, nside, lmax, outfile


@commands_hdf.command()
@click.argument("filename", type=click.STRING)
@click.argument("dataset", type=click.STRING)
def h52fits(
    filename,
    dataset,
):
    """
    Outputs a .h5 map to fits on the form 000001_cmb_amp_n1024.fits
    """
    import healpy as hp
    import h5py

    dataset, tag = dataset.rsplit("/", 1)

    with h5py.File(filename, "r") as f:
        maps = f[f"{dataset}/{tag}"][()]
        if "aml" in tag:
            lmax = f[f"{dataset}/amp_lmax"][()]  # Get lmax from h5

    nside = hp.npix2nside(maps.shape[-1])
    dataset = f"{dataset}/{tag}"
    outfile = dataset.replace("/", "_")
    outfile = outfile.replace("_map", "")
    hp.write_map(outfile + f"_n{str(nside)}.fits", maps, overwrite=True, dtype=None)


@commands_hdf.command()
@click.argument("input", type=click.STRING)
@click.argument("dataset", type=click.STRING)
@click.argument("nside", type=click.INT)
@click.option("-lmax", default=None, type=click.INT)
@click.option("-fwhm", default=0.0, type=click.FLOAT)
def alm2fits(input, dataset, nside, lmax, fwhm):
    """
    Converts c3 alms in .h5 file to fits.
    Specify nside and optional smoothing.
    """
    alm2fits_tool(input, dataset, nside, lmax, fwhm)


@commands_hdf.command()
@click.argument(
    "chain",
    type=click.Path(exists=True),
    nargs=-1,
)
@click.argument("burnin", type=click.INT)
@click.argument("procver", type=click.STRING)
@click.option(
    "-resamp",
    is_flag=True,
    help="data interpreted as resampled data",
)
@click.option(
    "-copy",
    "copy_",
    is_flag=True,
    help=" copy full .h5 file",
)
@click.option(
    "-freqmaps",
    is_flag=True,
    help=" output freqmaps",
)
@click.option(
    "-ame",
    is_flag=True,
    help=" output ame",
)
@click.option(
    "-ff",
    "-freefree",
    "ff",
    is_flag=True,
    help=" output freefree",
)
@click.option(
    "-cmb",
    is_flag=True,
    help=" output cmb",
)
@click.option(
    "-synch",
    is_flag=True,
    help=" output synchrotron",
)
@click.option(
    "-dust",
    is_flag=True,
    help=" output dust",
)
@click.option(
    "-cii",
    is_flag=True,
    help=" output dust cii",
)
@click.option(
    "-stars",
    is_flag=True,
    help=" output dust stars",
)
@click.option(
    "-hotpah",
    is_flag=True,
    help=" output hot PAH",
)
@click.option(
    "-co",
    is_flag=True,
    help=" output hot PAH",
)
@click.option(
    "-br",
    is_flag=True,
    help=" output BR",
)
@click.option("-diff", is_flag=True, help="Creates diff maps official releases")
@click.option("-diffcmb", is_flag=True, help="Creates diff maps cmb")
@click.option(
    "-goodness", is_flag=True, help="Output chisq and residual maps in separate dir"
)
@click.option("-chisq", is_flag=True, help="Output chisq ")
@click.option("-res", is_flag=True, help="Output residuals")
@click.option("-all", "all_", is_flag=True, help="Output all")
@click.option("-plot", is_flag=True, help="Plot everything (invoke plotrelease)")
@click.option("-pol", is_flag=True, help="if resamp is pol or T")
@click.pass_context
def release(
    ctx,
    chain,
    burnin,
    procver,
    resamp,
    copy_,
    freqmaps,
    ame,
    ff,
    cmb,
    synch,
    dust,
    co,
    cii,
    stars,
    hotpah,
    br,
    diff,
    diffcmb,
    goodness,
    chisq,
    res,
    all_,
    plot,
    pol,
):
    """
    Creates a release file-set on the BeyondPlanck format.
    https://gitlab.com/BeyondPlanck/repo/-/wikis/BeyondPlanck-Release-Candidate-2

    ex. c3pp release chains_v1_c{1,2}/chain_c000{1,2}.h5 30 CG_r1
    Will output formatted files using all chains specified,
    with a burnin of 30 to a directory called CG_r1

    This function outputs the following files to the {procver} directory:
    CG_chain01_{procver}.h5
    CG_resamp_chain01_Cl_{procver}.h5
    CG_resamp_chain01_noCl_{procver}.h5
    CG_param_v1.txt
    CG_param_resamp_Cl_v1.txt
    CG_param_resamp_noCl_v1.txt

    CG_030_IQU_n0512_{procver}.fits
    CG_044_IQU_n0512_{procver}.fits
    CG_070_IQU_n1024_{procver}.fits

    CG_cmb_IQU_n1024_{procver}.fits
    CG_synch_IQU_n1024_{procver}.fits
    CG_freefree_I_n1024_{procver}.fits
    CG_ame_I_n1024_{procver}.fits

    CG_cmb_GBRlike_{procver}.fits


    CG_DR2_{experiment_name}_{channel_id}_I_nside_{procver}
    """
    # TODO
    # Use proper masks for output of CMB component
    # Use inpainted data as well in CMB component

    from pathlib import Path
    import shutil

    if all_:  # sets all other flags to true
        copy_ = not copy_
        freqmaps = not freqmaps
        ame = not ame
        ff = not ff
        cmb = not cmb
        synch = not synch
        dust = not dust
        cii = not cii
        stars = not stars
        hotpah = not hotpah
        co = not co
        br = not br
        diff = not diff
        diffcmb = not diffcmb
        goodness = not goodness
        res = not res
        chisq = not chisq
        plot = not plot

    if goodness:
        chisq = res = True
    elif chisq or res:
        goodness = True

    # Make procver directory if not exists
    click.echo("{:#^80}".format(""))
    click.echo(f"Creating directory {procver}")
    Path(procver).mkdir(parents=True, exist_ok=True)
    chains = chain
    maxchain = len(chains)

    if pol:  # Dumb labeling thing
        pol = "P"
    else:
        pol = "T"
    """
    Copying chains files
    """
    if copy_:
        # Commander3 parameter file for main chain
        for i, chainfile in enumerate(chains, 1):
            path = os.path.split(chainfile)[0]
            for file in os.listdir(path):
                if file.startswith("param") and i == 1:  # Copy only first
                    click.echo(
                        f"Copying {path}/{file} to {procver}/CG_param_c"
                        + str(i).zfill(4)
                        + f"_{procver}.txt"
                    )
                    if resamp:
                        shutil.copyfile(
                            f"{path}/{file}",
                            f"{procver}/CG_param_c"
                            + str(i).zfill(4)
                            + f"_{pol}resamp_{procver}.txt",
                        )
                    else:
                        shutil.copyfile(
                            f"{path}/{file}",
                            f"{procver}/CG_param_c"
                            + str(i).zfill(4)
                            + f"_{procver}.txt",
                        )

            if resamp:
                # Resampled CMB-only full-mission Gibbs chain file with Cls (for BR estimator)
                click.echo(
                    f"Copying {chainfile} to {procver}/CG_c"
                    + str(i).zfill(4)
                    + f"_{pol}resamp_{procver}.h5"
                )
                shutil.copyfile(
                    chainfile,
                    f"{procver}/CG_c" + str(i).zfill(4) + f"_{pol}resamp_{procver}.h5",
                )
            else:
                # Full-mission Gibbs chain file
                click.echo(
                    f"Copying {chainfile} to {procver}/CG_c"
                    + str(i).zfill(4)
                    + f"_{procver}.h5"
                )
                shutil.copyfile(
                    chainfile,
                    f"{procver}/CG_c" + str(i).zfill(4) + f"_{procver}.h5",
                )

    # if halfring:
    #   # Copy halfring files
    #   for i, chainfile in enumerate([halfring], 1):
    #       # Copy halfring files
    #       click.echo(f"Copying {resamp} to {procver}/CG_halfring_c" + str(i).zfill(4) + f"_{procver}.h5")
    #       shutil.copyfile(halfring, f"{procver}/CG_halfring_c" + str(i).zfill(4) + f"_{procver}.h5",)

    """
    IQU mean, IQU stdev, (Masks for cmb)
    Run mean and stddev from min to max sample (Choose min manually or start at 1?)
    """
    if resamp:
        chain = f"{procver}/CG_c0001_{pol}resamp_{procver}.h5"
    else:
        chain = f"{procver}/CG_c0001_{procver}.h5"

    chain_bla = cosmoglobe.Chain(chain, burn_in=1)
    bands = chain_bla["000001/tod"]

    if freqmaps:
        try:
            # Ideally, we would have a way to access this information from the
            # chain itself, but currently we aren't outputting any of the main
            # stuff from the parameter file.
            # A lot of the nu_ref, bandctr, restfreq, bndwid parameters are just
            # dummies for now.

            
            if "023-WMAP_K" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="023-WMAP_K",
                    fwhm=0.0,
                    nu_ref_t="23 GHz",
                    nu_ref_p="23 GHz",
                    procver=procver,
                    filename=f"CG_023-WMAP_K_IQU_n0512_{procver}.fits",
                    bndctr=23,
                    restfreq=23,
                    bndwid=5,
                )

            if "030-WMAP_Ka" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="030-WMAP_Ka",
                    fwhm=0.0,
                    nu_ref_t="30 GHz",
                    nu_ref_p="30 GHz",
                    procver=procver,
                    filename=f"CG_030-WMAP_Ka_IQU_n0512_{procver}.fits",
                    bndctr=30,
                    restfreq=30,
                    bndwid=5,
                )
            if "040-WMAP_Q1" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="040-WMAP_Q1",
                    fwhm=0.0,
                    nu_ref_t="40 GHz",
                    nu_ref_p="40 GHz",
                    procver=procver,
                    filename=f"CG_040-WMAP_Q1_IQU_n0512_{procver}.fits",
                    bndctr=40,
                    restfreq=40,
                    bndwid=5,
                )
            if "040-WMAP_Q2" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="040-WMAP_Q2",
                    fwhm=0.0,
                    nu_ref_t="40 GHz",
                    nu_ref_p="40 GHz",
                    procver=procver,
                    filename=f"CG_040-WMAP_Q2_IQU_n0512_{procver}.fits",
                    bndctr=40,
                    restfreq=40,
                    bndwid=5,
                )
            if "060-WMAP_V1" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="060-WMAP_V1",
                    fwhm=0.0,
                    nu_ref_t="60 GHz",
                    nu_ref_p="60 GHz",
                    procver=procver,
                    filename=f"CG_060-WMAP_V1_IQU_n0512_{procver}.fits",
                    bndctr=60,
                    restfreq=60,
                    bndwid=5,
                )

            if "060-WMAP_V2" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="060-WMAP_V2",
                    fwhm=0.0,
                    nu_ref_t="60 GHz",
                    nu_ref_p="60 GHz",
                    procver=procver,
                    filename=f"CG_060-WMAP_V2_IQU_n0512_{procver}.fits",
                    bndctr=60,
                    restfreq=60,
                    bndwid=5,
                )

            if "090-WMAP_W1" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="090-WMAP_W1",
                    fwhm=0.0,
                    nu_ref_t="90 GHz",
                    nu_ref_p="90 GHz",
                    procver=procver,
                    filename=f"CG_090-WMAP_W1_IQU_n0512_{procver}.fits",
                    bndctr=90,
                    restfreq=90,
                    bndwid=5,
                )

            if "090-WMAP_W2" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="090-WMAP_W2",
                    fwhm=0.0,
                    nu_ref_t="90 GHz",
                    nu_ref_p="90 GHz",
                    procver=procver,
                    filename=f"CG_090-WMAP_W2_IQU_n0512_{procver}.fits",
                    bndctr=90,
                    restfreq=90,
                    bndwid=5,
                )
            if "090-WMAP_W3" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    nside=512,
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="090-WMAP_W3",
                    fwhm=0.0,
                    nu_ref_t="90 GHz",
                    nu_ref_p="90 GHz",
                    procver=procver,
                    filename=f"CG_090-WMAP_W3_IQU_n0512_{procver}.fits",
                    bndctr=90,
                    restfreq=90,
                    bndwid=5,
                )

            if "090-WMAP_W4" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    nside=512,
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                        "mK",
                        "mK",
                        "mK",
                        "mK2",
                    ],
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="090-WMAP_W4",
                    fwhm=0.0,
                    nu_ref_t="90 GHz",
                    nu_ref_p="90 GHz",
                    procver=procver,
                    filename=f"CG_090-WMAP_W4_IQU_n0512_{procver}.fits",
                    bndctr=90,
                    restfreq=90,
                    bndwid=5,
                )

            # Full-mission 30 GHz IQU frequency map
            # CG_030_IQU_n0512_{procver}.fits
            if "030" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK2",
                        "uK",
                        "uK",
                        "uK",
                        "uK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="030",
                    fwhm=0.0,
                    nu_ref_t="30.0 GHz",
                    nu_ref_p="30.0 GHz",
                    procver=procver,
                    filename=f"CG_030_IQU_n0512_{procver}.fits",
                    bndctr=30,
                    restfreq=28.456,
                    bndwid=9.899,
                )
            # Full-mission 44 GHz IQU frequency map
            if "044" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK2",
                        "uK",
                        "uK",
                        "uK",
                        "uK2",
                    ],
                    nside=512,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="044",
                    fwhm=0.0,
                    nu_ref_t="44.0 GHz",
                    nu_ref_p="44.0 GHz",
                    procver=procver,
                    filename=f"CG_044_IQU_n0512_{procver}.fits",
                    bndctr=44,
                    restfreq=44.121,
                    bndwid=10.719,
                )
            
            # Full-mission 70 GHz IQU frequency map
            if "070" in bands:
                format_fits(
                    chain=chain,
                    extname="FREQMAP",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_RMS",
                        "Q_RMS",
                        "U_RMS",
                        "QU_RMS",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "QU_COV",
                    ],
                    units=[
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK",
                        "uK2",
                        "uK",
                        "uK",
                        "uK",
                        "uK2",
                    ],
                    nside=1024,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="070",
                    fwhm=0.0,
                    nu_ref_t="70.0 GHz",
                    nu_ref_p="70.0 GHz",
                    procver=procver,
                    filename=f"CG_070_IQU_n1024_{procver}.fits",
                    bndctr=70,
                    restfreq=70.467,
                    bndwid=14.909,
                )

            # DIRBE bands
            for b in range(1, 11):
                import astropy.units as u
                import astropy.constants as c
                wavs = np.array([1.25, 2.2, 3.5, 4.9, 12, 25, 60, 100, 140, 240])*u.micron
                bw = np.array([59.5, 22.4, 22.0, 8.19, 13.3, 4.13, 2.32, 0.974, 0.605, 0.495])*u.THz
                bw = bw.to('GHz')
                freqs = (c.c/wavs).to('GHz')
                if (f"{b:02}a" in bands) & (f"{b:02}b" in bands):

                    band_cent = int(freqs[b-1].value)
                    bandwidth = int(bw[b-1].value)

                    format_fits(
                        chain=chain,
                        extname="FREQMAP",
                        types=[
                            "I_MEAN",
                            "I_RMS",
                            "I_STDDEV",
                        ],
                        units=[
                            "MJy/sr",
                            "MJy/sr",
                            "MJy/sr",
                        ],
                        nside=512,
                        burnin=burnin,
                        maxchain=maxchain,
                        polar=False,
                        component=[f"{b:02}a", f"{b:02}b", f"{b:02}"],
                        fwhm=0.0,
                        nu_ref_t=f'{band_cent} GHz',
                        nu_ref_p=None,
                        procver=procver,
                        filename=f"CG_DIRBE_{b:02}_I_n0512_{procver}.fits",
                        bndctr=band_cent,
                        restfreq=band_cent,
                        bndwid=bandwidth,
                        coadd=True,
                    )

                    format_fits(
                        chain=chain,
                        extname="FREQMAP",
                        types=[
                            "I_MEAN",
                            "I_RMS",
                            "I_STDDEV",
                        ],
                        units=[
                            "MJy/sr",
                            "MJy/sr",
                            "MJy/sr",
                        ],
                        nside=512,
                        burnin=burnin,
                        maxchain=maxchain,
                        polar=False,
                        component=f"{b:02}a",
                        fwhm=0.0,
                        nu_ref_t=f'{band_cent} GHz',
                        nu_ref_p=None,
                        procver=procver,
                        filename=f"CG_DIRBE_{b:02}a_I_n0512_{procver}.fits",
                        bndctr=band_cent,
                        restfreq=band_cent,
                        bndwid=bandwidth,
                    )

                    format_fits(
                        chain=chain,
                        extname="FREQMAP",
                        types=[
                            "I_MEAN",
                            "I_RMS",
                            "I_STDDEV",
                        ],
                        units=[
                            "MJy/sr",
                            "MJy/sr",
                            "MJy/sr",
                        ],
                        nside=512,
                        burnin=burnin,
                        maxchain=maxchain,
                        polar=False,
                        component=f"{b:02}b",
                        fwhm=0.0,
                        nu_ref_t=f'{band_cent} GHz',
                        nu_ref_p=None,
                        procver=procver,
                        filename=f"CG_DIRBE_{b:02}b_I_n0512_{procver}.fits",
                        bndctr=band_cent,
                        restfreq=band_cent,
                        bndwid=bandwidth,
                    )


        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    """
    FOREGROUND MAPS
    """
    # Full-mission CMB IQU map
    if cmb:
        if resamp:
            if pol == "P":
                try:
                    format_fits(
                        chain,
                        extname="COMP-MAP-CMB-RESAMP-P",
                        types=[
                            "Q_MEAN",
                            "U_MEAN",
                            "Q_STDDEV",
                            "U_STDDEV",
                        ],
                        units=[
                            "uK_cmb",
                            "uK_cmb",
                            "uK_cmb",
                            "uK_cmb",
                        ],
                        nside=1024,
                        burnin=burnin,
                        maxchain=maxchain,
                        polar=True,
                        component="CMB",
                        fwhm=14.0,
                        nu_ref_t="NONE",
                        nu_ref_p="NONE",
                        procver=procver,
                        filename=f"CG_cmb_resamp_QU_n1024_{procver}.fits",
                        bndctr=None,
                        restfreq=None,
                        bndwid=None,
                    )
                except Exception as e:
                    print(e)
                    click.secho("Continuing...", fg="yellow")

            else:
                try:
                    format_fits(
                        chain,
                        extname="COMP-MAP-CMB-RESAMP-T",
                        types=[
                            "I_MEAN",
                            "I_STDDEV",
                        ],
                        units=[
                            "uK_cmb",
                            "uK_cmb",
                        ],
                        nside=1024,
                        burnin=burnin,
                        maxchain=maxchain,
                        polar=True,
                        component="CMB",
                        fwhm=14.0,
                        nu_ref_t="NONE",
                        nu_ref_p="NONE",
                        procver=procver,
                        filename=f"CG_cmb_resamp_I_n1024_{procver}.fits",
                        bndctr=None,
                        restfreq=None,
                        bndwid=None,
                    )
                except Exception as e:
                    print(e)
                    click.secho("Continuing...", fg="yellow")

        else:
            try:
                format_fits(
                    chain,
                    extname="COMP-MAP-CMB",
                    types=[
                        "I_MEAN",
                        "Q_MEAN",
                        "U_MEAN",
                        "I_STDDEV",
                        "Q_STDDEV",
                        "U_STDDEV",
                        "mask1",
                        "mask2",
                    ],
                    units=[
                        "uK_cmb",
                        "uK_cmb",
                        "uK",
                        "uK",
                        "NONE",
                        "NONE",
                    ],
                    nside=1024,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="CMB",
                    fwhm=14.0,
                    nu_ref_t="NONE",
                    nu_ref_p="NONE",
                    procver=procver,
                    filename=f"CG_cmb_IQU_n1024_{procver}.fits",
                    bndctr=None,
                    restfreq=None,
                    bndwid=None,
                )
            except Exception as e:
                print(e)
                click.secho("Continuing...", fg="yellow")

    if ff:
        try:
            # Full-mission free-free I map
            format_fits(
                chain,
                extname="COMP-MAP-FREE-FREE",
                types=[
                    "I_MEAN",
                    "I_TE_MEAN",
                    "I_STDDEV",
                    "I_TE_STDDEV",
                ],
                units=[
                    "uK_RJ",
                    "K",
                    "uK_RJ",
                    "K",
                ],
                nside=1024,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="FREE-FREE",
                fwhm=30.0,
                nu_ref_t="40.0 GHz",
                nu_ref_p="40.0 GHz",
                procver=procver,
                filename=f"CG_freefree_I_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if ame:
        try:
            # Full-mission AME I map
            format_fits(
                chain,
                extname="COMP-MAP-AME",
                types=[
                    "I_MEAN",
                    "I_NU_P_MEAN",
                    "I_STDDEV",
                    "I_NU_P_STDDEV",
                ],
                units=[
                    "uK_RJ",
                    "GHz",
                    "uK_RJ",
                    "GHz",
                ],
                nside=1024,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="AME",
                fwhm=120.0,
                nu_ref_t="22.0 GHz",
                nu_ref_p="22.0 GHz",
                procver=procver,
                filename=f"CG_ame_I_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if synch:
        try:
            # Full-mission synchrotron IQU map
            format_fits(
                chain,
                extname="COMP-MAP-SYNCHROTRON",
                types=[
                    "I_MEAN",
                    "Q_MEAN",
                    "U_MEAN",
                    "P_MEAN",
                    "I_BETA_MEAN",
                    "QU_BETA_MEAN",
                    "I_STDDEV",
                    "Q_STDDEV",
                    "U_STDDEV",
                    "P_STDDEV",
                    "I_BETA_STDDEV",
                    "QU_BETA_STDDEV",
                ],
                units=[
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "NONE",
                    "NONE",
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "NONE",
                    "NONE",
                ],
                nside=1024,
                burnin=burnin,
                maxchain=maxchain,
                polar=True,
                component="SYNCHROTRON",
                fwhm=60.0,  # 60.0,
                nu_ref_t="0.408 GHz",
                nu_ref_p="30.0 GHz",
                procver=procver,
                filename=f"CG_synch_IQU_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if dust:
        # Need a better way to deal with the temperature-only analyses.
        try:
            # Full-mission thermal dust IQU map
            format_fits(
                chain,
                extname="COMP-MAP-DUST",
                types=[
                    "I_MEAN",
                    "I_BETA_MEAN",
                    "I_T_MEAN",
                    "I_STDDEV",
                    "I_BETA_STDDEV",
                    "I_T_STDDEV",
                ],
                units=[
                    "uK_RJ",
                    "NONE",
                    "K",
                    "uK_RJ",
                    "NONE",
                    "K",
                ],
                nside=2048,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="DUST",
                fwhm=10.0,  # 60.0,
                nu_ref_t="545 GHz",
                nu_ref_p=None,
                procver=procver,
                filename=f"CG_dust_I_n2048_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

        try:
            # Full-mission thermal dust IQU map
            format_fits(
                chain,
                extname="COMP-MAP-DUST",
                types=[
                    "I_MEAN",
                    "Q_MEAN",
                    "U_MEAN",
                    "P_MEAN",
                    "I_BETA_MEAN",
                    "QU_BETA_MEAN",
                    "I_T_MEAN",
                    "QU_T_MEAN",
                    "I_STDDEV",
                    "Q_STDDEV",
                    "U_STDDEV",
                    "P_STDDEV",
                    "I_BETA_STDDEV",
                    "QU_BETA_STDDEV",
                    "I_T_STDDEV",
                    "QU_T_STDDEV",
                ],
                units=[
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "NONE",
                    "NONE",
                    "K",
                    "K",
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "uK_RJ",
                    "NONE",
                    "NONE",
                    "K",
                    "K",
                ],
                nside=1024,
                burnin=burnin,
                maxchain=maxchain,
                polar=True,
                component="DUST",
                fwhm=10.0,  # 60.0,
                nu_ref_t="545 GHz",
                nu_ref_p="353 GHz",
                procver=procver,
                filename=f"CG_dust_IQU_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")


    if cii:
        try:
            # Full-mission free-free I map
            format_fits(
                chain,
                extname="COMP-MAP-CII-line",
                types=[
                    "I_MEAN",
                    "I_STDDEV",
                ],
                units=[
                    "MJy/sr",
                    "MJy/sr",
                ],
                nside=2048,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="CII",
                fwhm=10.0,
                nu_ref_t="40.0 GHz",
                nu_ref_p="40.0 GHz",
                procver=procver,
                filename=f"CG_cii_I_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if stars:
        # Stars as evaluated at DIRBE 01a
        try:
            if len(chains) == 1:
                cmin = 1
                cmax = None
                chdir = os.path.split(chains[0])[0].rsplit("chain_", 1)[0]
            else:
                cmin = 1
                cmax = len(chains)
                chdir = os.path.split(chains[0])[0]
            format_fits(
                chain,
                extname="COMP-MAP-stars",
                types=[
                    "I_MEAN",
                    "I_STDDEV",
                ],
                units=[
                    "MJy/sr",
                    "MJy/sr",
                ],
                nside=512,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="CII",
                fwhm=10.0,
                nu_ref_t="40.0 GHz",
                nu_ref_p="40.0 GHz",
                procver=procver,
                filename=f"CG_stars_I_n0512_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
                chdir=chdir,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if hotpah:
        try:
            # Full-mission free-free I map
            format_fits(
                chain,
                extname="COMP-hotPAH",
                types=[
                    "I_MEAN",
                    "I_STDDEV",
                ],
                units=[
                    "MJy/sr",
                    "MJy/sr",
                ],
                nside=2048,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="hotPAH",
                fwhm=10.0,
                nu_ref_t="40.0 GHz",
                nu_ref_p="40.0 GHz",
                procver=procver,
                filename=f"CG_hotPAH_I_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")
    if co:
        try:
            # Full-mission free-free I map
            format_fits(
                chain,
                extname="COMP-CO_tot",
                types=[
                    "I_MEAN",
                    "I_STDDEV",
                ],
                units=[
                    "MJy/sr",
                    "MJy/sr",
                ],
                nside=1024,
                burnin=burnin,
                maxchain=maxchain,
                polar=False,
                component="co_tot",
                fwhm=30.0,
                nu_ref_t="40.0 GHz",
                nu_ref_p="40.0 GHz",
                procver=procver,
                filename=f"CG_CO_tot_I_n1024_{procver}.fits",
                bndctr=None,
                restfreq=None,
                bndwid=None,
            )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if diff:
        import healpy as hp

        try:
            if not os.path.exists(f"{procver}/diffs"):
                os.mkdir(f"{procver}/diffs")
            click.echo("Creating frequency difference maps")
            path_dx12 = "/mn/stornext/u3/trygvels/compsep/cdata/like/BP_releases/dx12"
            path_npipe = "/mn/stornext/u3/trygvels/compsep/cdata/like/BP_releases/npipe"
            path_BP10 = "/mn/stornext/d16/cmbco/bp/delivery/v10.00/v2"
            maps_dx12 = [
                "30ghz_2018_n1024_beamscaled_dip.fits",
                "44ghz_2018_n1024_beamscaled_dip.fits",
                "70ghz_2018_n1024_beamscaled_dip.fits",
            ]
            maps_npipe = [
                "npipe6v20_030_map_uK.fits",
                "npipe6v20_044_map_uK.fits",
                "npipe6v20_070_map_uK.fits",
            ]
            maps_BP10 = [
                "BP_030_IQU_n0512_v2.fits",
                "BP_044_IQU_n0512_v2.fits",
                "BP_070_IQU_n1024_v2.fits",
            ]
            maps_CG = [
                f"CG_030_IQU_n0512_{procver}.fits",
                f"CG_044_IQU_n0512_{procver}.fits",
                f"CG_070_IQU_n1024_{procver}.fits",
            ]
            beamscaling = [9.8961854e-01, 9.9757886e-01, 9.9113965e-01]

            for i, freq in enumerate(
                [
                    "030",
                    "044",
                    "070",
                ]
            ):
                map_CG = hp.read_map(
                    f"{procver}/{maps_CG[i]}", field=(0, 1, 2), dtype=None
                )
                map_npipe = hp.read_map(
                    f"{path_npipe}/{maps_npipe[i]}", field=(0, 1, 2), dtype=None
                )
                map_dx12 = hp.read_map(
                    f"{path_dx12}/{maps_dx12[i]}", field=(0, 1, 2), dtype=None
                )
                map_BP10 = hp.read_map(
                    f"{path_BP10}/{maps_BP10[i]}", field=(0, 1, 2), dtype=None
                )

                # dx12 dipole values:
                # 3362.08 pm 0.99, 264.021 pm 0.011, 48.253 ± 0.005
                # 233.18308357  2226.43833645 -2508.42179665
                # dipole_dx12 = -3362.08*hp.dir2vec(264.021, 48.253, lonlat=True)

                # map_dx12  = map_dx12/beamscaling[i]
                # Smooth to 60 arcmin
                map_CG = hp.smoothing(map_CG, fwhm=arcmin2rad(60.0))
                map_npipe = hp.smoothing(map_npipe, fwhm=arcmin2rad(60.0))
                map_dx12 = hp.smoothing(map_dx12, fwhm=arcmin2rad(60.0))
                map_BP10 = hp.smoothing(map_BP10, fwhm=arcmin2rad(60.0))

                # ud_grade 30 and 44ghz
                if i < 2:
                    map_npipe = hp.ud_grade(
                        map_npipe,
                        nside_out=512,
                    )
                    map_dx12 = hp.ud_grade(
                        map_dx12,
                        nside_out=512,
                    )
                    map_BP10 = hp.ud_grade(
                        map_BP10,
                        nside_out=512,
                    )

                # Remove monopoles
                map_CG -= np.mean(map_CG, axis=1).reshape(-1, 1)
                map_npipe -= np.mean(map_npipe, axis=1).reshape(-1, 1)
                map_dx12 -= np.mean(map_dx12, axis=1).reshape(-1, 1)
                map_BP10 -= np.mean(map_BP10, axis=1).reshape(-1, 1)
                click.echo(f"creating {freq} GHz difference")
                hp.write_map(
                    f"{procver}/diffs/CG_{freq}_diff_npipe_{procver}.fits",
                    np.array(map_CG - map_npipe),
                    overwrite=True,
                    column_names=["I_DIFF", "Q_DIFF", "U_DIFF"],
                    dtype=None,
                )
                hp.write_map(
                    f"{procver}/diffs/CG_{freq}_diff_dx12_{procver}.fits",
                    np.array(map_CG - map_dx12),
                    overwrite=True,
                    column_names=["I_DIFF", "Q_DIFF", "U_DIFF"],
                    dtype=None,
                )
                hp.write_map(
                    f"{procver}/diffs/CG_{freq}_diff_BP10_{procver}.fits",
                    np.array(map_CG - map_BP10),
                    overwrite=True,
                    column_names=["I_DIFF", "Q_DIFF", "U_DIFF"],
                    dtype=None,
                )

            path_wmap9 = "/mn/stornext/d16/cmbco/ola/wmap/freq_maps"
            maps_wmap9 = [
                "wmap_iqusmap_r9_9yr_Ka1_v5.fits",
                "wmap_iqusmap_r9_9yr_Q1_v5.fits",
                "wmap_iqusmap_r9_9yr_Q2_v5.fits",
                "wmap_iqusmap_r9_9yr_V1_v5.fits",
                "wmap_iqusmap_r9_9yr_V2_v5.fits",
                "wmap_iqusmap_r9_9yr_K1_v5.fits",
                "wmap_iqusmap_r9_9yr_W1_v5.fits",
                "wmap_iqusmap_r9_9yr_W2_v5.fits",
                "wmap_iqusmap_r9_9yr_W3_v5.fits",
                "wmap_iqusmap_r9_9yr_W4_v5.fits",
            ]
            maps_CG = [
                f"CG_030-WMAP_Ka_IQU_n0512_{procver}.fits",
                f"CG_040-WMAP_Q1_IQU_n0512_{procver}.fits",
                f"CG_040-WMAP_Q2_IQU_n0512_{procver}.fits",
                f"CG_060-WMAP_V1_IQU_n0512_{procver}.fits",
                f"CG_060-WMAP_V2_IQU_n0512_{procver}.fits",
                f"CG_023-WMAP_K_IQU_n0512_{procver}.fits",
                f"CG_090-WMAP_W1_IQU_n0512_{procver}.fits",
                f"CG_090-WMAP_W2_IQU_n0512_{procver}.fits",
                f"CG_090-WMAP_W3_IQU_n0512_{procver}.fits",
                f"CG_090-WMAP_W4_IQU_n0512_{procver}.fits",
            ]

            # WMAP9 maps must have dipole added back in:
            # Jarosik 2011 gives (unchanged from 5-year release)
            # d = 3.355, l = 263.99, b = 48.26
            d_x = -0.233
            d_y = -2.222
            d_z = 2.504
            import healpy as hp

            x, y, z = hp.pix2vec(512, np.arange(12 * 512**2))
            dip = d_x * x + d_y * y + d_z * z
            for i, freq in enumerate(
                [
                    "023-WMAP_K",
                    "030-WMAP_Ka",
                    "040-WMAP_Q1",
                    "040-WMAP_Q2",
                    "060-WMAP_V1",
                    "060-WMAP_V2",
                    "090-WMAP_W1",
                    "090-WMAP_W2",
                    "090-WMAP_W3",
                    "090-WMAP_W4",
                ]
            ):
                map_CG = hp.read_map(
                    f"{procver}/{maps_CG[i]}", field=(0, 1, 2), dtype=None
                )
                map_wmap9 = hp.read_map(
                    f"{path_wmap9}/{maps_wmap9[i]}", field=(0, 1, 2), dtype=None
                )

                map_wmap9[0] += dip

                # map_dx12  = map_dx12/beamscaling[i]
                # Smooth to 60 arcmin
                map_CG = hp.smoothing(map_CG, fwhm=arcmin2rad(60.0))
                map_wmap9 = hp.smoothing(map_wmap9, fwhm=arcmin2rad(60.0))

                # Remove monopoles
                map_CG -= np.mean(map_CG, axis=1).reshape(-1, 1)
                map_wmap9 -= np.mean(map_wmap9, axis=1).reshape(-1, 1)
                click.echo(f"creating {freq} GHz difference")
                hp.write_map(
                    f"{procver}/diffs/CG_{freq}_diff_wmap9_{procver}.fits",
                    np.array(map_CG - map_wmap9),
                    overwrite=True,
                    column_names=["I_DIFF", "Q_DIFF", "U_DIFF"],
                    dtype=None,
                )
        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if diffcmb:
        import healpy as hp

        try:
            if not os.path.exists(f"{procver}/diffs"):
                os.mkdir(f"{procver}/diffs")
            click.echo("Creating cmb difference maps")
            path_cmblegacy = (
                "/mn/stornext/u3/trygvels/compsep/cdata/like/BP_releases/cmb-legacy"
            )
            mask_ = hp.read_map(
                "/mn/stornext/u3/trygvels/compsep/cdata/like/BP_releases/masks/dx12_v3_common_mask_int_005a_1024_TQU.fits",
                dtype=np.bool,
            )
            map_CG = hp.read_map(
                f"{procver}/CG_cmb_IQU_n1024_{procver}.fits",
                field=(0, 1, 2),
                dtype=None,
            )
            map_CG_masked = hp.ma(map_CG[0])
            map_CG_masked.mask = np.logical_not(mask_)
            mono, dip = hp.fit_dipole(map_CG_masked)
            nside = 1024
            ray = range(hp.nside2npix(nside))
            vecs = hp.pix2vec(nside, ray)
            dipole = np.dot(dip, vecs)
            map_CG[0] = map_CG[0] - dipole - mono
            map_CG = hp.smoothing(map_CG, fwhm=arcmin2rad(np.sqrt(60.0**2 - 14**2)))
            # map_CG -= np.mean(map_CG,axis=1).reshape(-1,1)
            for i, method in enumerate(
                [
                    "commander",
                    "sevem",
                    "nilc",
                    "smica",
                ]
            ):

                data = f"COM_CMB_IQU-{method}_2048_R3.00_full.fits"
                click.echo(f"making difference map with {data}")
                map_cmblegacy = hp.read_map(
                    f"{path_cmblegacy}/{data}", field=(0, 1, 2), dtype=None
                )
                map_cmblegacy = hp.smoothing(map_cmblegacy, fwhm=arcmin2rad(60.0))
                map_cmblegacy = hp.ud_grade(
                    map_cmblegacy,
                    nside_out=1024,
                )
                map_cmblegacy = map_cmblegacy * 1e6

                # Remove monopoles
                map_cmblegacy_masked = hp.ma(map_cmblegacy[0])
                map_cmblegacy_masked.mask = np.logical_not(mask_)
                mono = hp.fit_monopole(map_cmblegacy_masked)
                click.echo(f"{method} subtracting monopole {mono}")
                map_cmblegacy[0] = (
                    map_cmblegacy[0] - mono
                )  # np.mean(map_cmblegacy,axis=1).reshape(-1,1)

                hp.write_map(
                    f"{procver}/diffs/CG_cmb_diff_{method}_{procver}.fits",
                    np.array(map_CG - map_cmblegacy),
                    overwrite=True,
                    column_names=["I_DIFF", "Q_DIFF", "U_DIFF"],
                    dtype=None,
                )

        except Exception as e:
            print(e)
            click.secho("Continuing...", fg="yellow")

    if goodness:
        import healpy as hp

        path_goodness = procver + "/goodness"
        Path(path_goodness).mkdir(parents=True, exist_ok=True)
        print("PATH", path_goodness)

        if len(chains) == 1:
            cmin = 1
            cmax = None
            chdir = os.path.split(chains[0])[0].rsplit("chain_", 1)[0]
        else:
            #print(chains[0])
            #print(os.path.split(chains[0]))
            #print(os.path.split(chains[0])[-1])
            #print(os.path.split(chains[0])[-1][-1])
            #cmin = int(os.path.split(chains[0])[0].rsplit("_c")[-1])
            #cmax = int(os.path.split(chains[-1])[0].rsplit("_c")[-1])
            cmin = 1
            cmax = len(chains)
            chdir = os.path.split(chains[0])[0]
            #chdir = os.path.split(chains[0])[0].rsplit("_", 1)[0]
            #print(chdir)

        if chisq:
            try:
                format_fits(
                    chains,
                    extname="CHISQ",
                    types=[
                        "I_MEAN",
                        "P_MEAN",
                    ],
                    units=[
                        "NONE",
                        "NONE",
                    ],
                    nside=16,
                    burnin=burnin,
                    maxchain=maxchain,
                    polar=True,
                    component="CHISQ",
                    fwhm=0.0,
                    nu_ref_t="NONE",
                    nu_ref_p="NONE",
                    procver=procver,
                    filename=f"goodness/CG_chisq_n16_{procver}.fits",
                    bndctr=None,
                    restfreq=None,
                    bndwid=None,
                    cmin=cmin,
                    cmax=cmax,
                    chdir=chdir,
                )
            except Exception as e:
                print(e)
                click.secho("Continuing...", fg="yellow")

        if res:
            click.echo("Save and format chisq map and residual maps")
            bands = {
                "030": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "uK",
                    "scale": 1.0,
                    "coadd": False
                },
                "044": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "uK",
                    "scale": 1.0,
                    "coadd": False
                },
                "070": {
                    "nside": 1024,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "uK",
                    "scale": 1.0,
                    "coadd": False
                },
                "023-WMAP_K": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1,
                    "coadd": False
                },
                "030-WMAP_Ka": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1,
                    "coadd": False
                },
                "040-WMAP_Q1": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "040-WMAP_Q2": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (
                        0,
                        1,
                        2,
                    ),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "060-WMAP_V1": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "060-WMAP_V2": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "090-WMAP_W1": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "090-WMAP_W2": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "090-WMAP_W3": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "090-WMAP_W4": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "IQU",
                    "fields": (0, 1, 2),
                    "unit": "mK",
                    "scale": 1.0,
                    "coadd": False
                },
                "0.4-Haslam": {
                    "nside": 512,
                    "fwhm": 120,
                    "sig": "I",
                    "fields": (0,),
                    "unit": "uK",
                    "scale": 1.0,
                    "coadd": False
                },
                #"857": {
                #    "nside": 1024,
                #    "fwhm": 120,
                #    "sig": "I",
                #    "fields": (0,),
                #    "unit": "uK",
                #    "scale": 1.0,
                #},
                #"353": {
                #    "nside": 1024,
                #    "fwhm": 120,
                #    "sig": "QU",
                #    "fields": (1, 2),
                #    "unit": "uK",
                #    "scale": 1.0,
                #},
            }

            for i in range(1, 11):
                # DIRBE bands
                bands[f"{i:02}a"] =  {
                    "nside": 512,
                    "fwhm": 0,
                    "sig": "I",
                    "fields": (0,),
                    "unit": "MJy/sr",
                    "scale": 1.0,
                    "coadd": False
                }
                bands[f"{i:02}b"] =  {
                    "nside": 512,
                    "fwhm": 0,
                    "sig": "I",
                    "fields": (0,),
                    "unit": "MJy/sr",
                    "scale": 1.0,
                    "coadd": False
                }
                bands[f"{i:02}"] =  {
                    "nside": 512,
                    "fwhm": 0,
                    "sig": "I",
                    "fields": (0,),
                    "unit": "MJy/sr",
                    "scale": 1.0,
                    "coadd": True
                }

            for lab, b in bands.items():
                print(lab, b)
                label = lab

                types = []
                units = []
                for l in b["sig"]:
                    types.append(f"{l}_MEAN")
                    units.append(b["unit"])
                for l in b["sig"]:
                    types.append(f"{l}_STDDEV")
                    units.append(b["unit"])
                try:
                    if b['coadd']:
                        lab = [f'{lab}a', f'{lab}b', f'{lab}']
                    else:
                        lab = lab
                    format_fits(
                        chains,
                        extname="FREQBAND_RES",
                        types=types,
                        units=units,
                        nside=b["nside"],
                        burnin=burnin,
                        maxchain=maxchain,
                        polar=len(b["fields"]) > 1,
                        component=lab,
                        fwhm=b["fwhm"],
                        nu_ref_t="NONE",
                        nu_ref_p="NONE",
                        procver=procver,
                        filename=f'goodness/CG_res_{label}_{b["sig"]}_n{b["nside"]}_{b["fwhm"]}arcmin_{procver}.fits',
                        bndctr=None,
                        restfreq=None,
                        bndwid=None,
                        cmin=cmin,
                        cmax=cmax,
                        chdir=chdir,
                        fields=b["fields"],
                        scale=b["scale"],
                        coadd=b['coadd'],
                    )
                except Exception as e:
                    print(e)
                    click.secho("Continuing...", fg="yellow")

    """ As implemented by Simone
    """
    if br and resamp:
        # Gaussianized TT Blackwell-Rao input file
        click.echo("{:-^50}".format("CMB GBR"))
        ctx.invoke(
            sigma_l2fits,
            filename=resamp,
            nchains=1,
            burnin=burnin,
            path="cmb/sigma_l",
            outname=f"{procver}/CG_cmb_GBRlike_{procver}.fits",
            save=True,
        )

    """
    TODO Generalize this so that they can be generated by Elina and Anna-Stiina
    """
    # Full-mission 30 GHz IQU beam symmetrized frequency map
    # CG_030_IQUdeconv_n0512_{procver}.fits
    # Full-mission 44 GHz IQU beam symmetrized frequency map
    # CG_044_IQUdeconv_n0512_{procver}.fits
    # Full-mission 70 GHz IQU beam symmetrized frequency map
    # CG_070_IQUdeconv_n1024_{procver}.fits

    """ Both sigma_l's and Dl's re in the h5. (Which one do we use?)
    """
    # CMB TT, TE, EE power spectrum
    # CG_cmb_{procver}.txt

    """ Just get this from somewhere
    """
    # Best-fit LCDM CMB TT, TE, EE power spectrum
    # CG_cmb_bfLCDM_{procver}.txt

    if plot:
        os.chdir(procver)
        ctx.invoke(plotrelease, procver=procver, all_=True)



# ================================================================================
#
# Copyright (C) 2020 Institute of Theoretical Astrophysics, University of Oslo.
#
# This file is part of Commander3.
#
# Commander3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Commander3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Commander3. If not, see <https://www.gnu.org/licenses/>.
#
# ================================================================================


import h5py
from cosmoglobe.tod_tools import huffman, rice
import numpy as np
import os


class TODLoader:
    def __init__(self, outPath, name, version=None, dicts=None, overwrite=False):
        self.outPath = outPath
        self.filelists = dicts
        self.version = version
        self.name = name
        # TODO: something with the version number
        self.overwrite = overwrite

    # initilizes a file for a single od
    def init_file(self, freq, od, mode="r"):
        self.huffDict = {}
        self.raggedDict = {}
        self.attrDict = {}
        self.encodings = {}
        self.pids = {}

        self.od = od
        self.freq = freq
        if self.name.lower() == "planck":
            sfreq = str(freq).zfill(3)
        else:
            sfreq = str(freq)
        if not self.od:
            if not self.name:
                self.outName = os.path.join(self.outPath, sfreq + ".h5")
            else:
                self.outName = os.path.join(
                    self.outPath, self.name + "_" + sfreq + ".h5"
                #self.outName = os.path.join(
                #    self.outPath, self.name + sfreq + ".h5"
                )
        else:
            self.outName = os.path.join(
                self.outPath, self.name + "_" + sfreq + "_" + str(od).zfill(6) + ".h5"
            )

        self.exists = False

        if os.path.exists(self.outName):
            self.exists = True
        if mode == "w":
            if self.exists and self.overwrite:
                os.remove(self.outName)
            try:
                self.outFile = h5py.File(self.outName, "a")

                if self.exists and not self.overwrite:
                    for pid in self.load_field("/common/pids"):
                        loadBalance = self.load_field(
                            "/" + str(pid).zfill(6) + "/common/load"
                        )
                        self.pids[pid] = (
                            str(float(loadBalance[0]))
                            + " "
                            + str(float(loadBalance[1]))
                        )
            except (KeyError, OSError):
                if hasattr(self, "outFile"):
                    self.outFile.close()
                os.remove(self.outName)
                self.exists = False
                self.outFile = h5py.File(self.outName, "a")

        if mode == "r":
            if not self.exists:
                raise OSError("Cannot find file " + self.outName)
            self.outFile = h5py.File(self.outName, "r")

    # Write a possibly compressed, possibly ragged matrix
    def add_matrix(self, fieldName, data, columnInfo, compression=None):

        writeField = True
        dims = np.shape(data)

        if len(dims) != 2:
            raise TypeError("Call to add matrix with an object of shape " + str(shape))

        if dims[0] != len(columnInfo):
            if dims[1] == len(columnInfo):
                data = np.transpose(data)
            else:
                raise ValueError(
                    "Data is shape "
                    + str(shape)
                    + " but column headers have length "
                    + str(len(columnInfo))
                )

        if compression == None or compression == []:
            if type(data) == np.ndarray:
                self.outFile.create_dataset(fieldName, data=data)
            else:  # ragged array
                dt = h5py.vlen_dtype(np.dtype(data[0]))
                dset = f.create_dataset(fieldName, dims[0], dtype=dt)
                for i, vec in zip(range(len(data)), data):
                    dset[i] = data
        else:  # compressed array
            for i, vec in zip(range(len(data)), data):
                self.add_field(fieldName + ":" + str(i), data[i], compression)

        self.add_attribute(fieldName, "index", columnInfo)
        self.add_attribute(fieldName, "matrix", True)

    # Single field write
    def add_field(self, fieldName, data, compression=None):
        data = np.nan_to_num(data)
        writeField = True
        if compression is not None and compression is not []:
            compInfo = ""
            if len(compression) == 2 and type(compression[0]) == str:
                # catch case with only one compression argument not in array
                compression = [compression]
            for compArr in compression:
                compInfo += compArr[0] + " "
                if compArr[0] == "dtype":
                    data = np.array(data, dtype=compArr[1]["dtype"])

                elif compArr[0] == "sigma":
                    data = np.int32(
                        compArr[1]["nsigma"]
                        * (data - compArr[1]["offset"])
                        / (compArr[1]["sigma0"])
                    )
                    metaName = "/common/n" + fieldName.split("/")[-1] + "sigma"
                    self.encodings[metaName] = compArr[1]["nsigma"]

                    self.add_attribute(fieldName, "nsigma", compArr[1]["nsigma"])
                    self.add_attribute(fieldName, "sigma0", compArr[1]["sigma0"])
                    self.add_attribute(fieldName, "offset", compArr[1]["offset"])

                elif compArr[0] == "digitize":
                    bins = np.linspace(
                        compArr[1]["min"], compArr[1]["max"], num=compArr[1]["nbins"]
                    )
                    data = np.digitize(data, bins)
                    metaName = "/common/n" + fieldName.split("/")[-1]
                    self.add_encoding(metaName, compArr[1]["nbins"])
                    self.add_attribute(fieldName, "min", compArr[1]["min"])
                    self.add_attribute(fieldName, "max", compArr[1]["max"])
                    self.add_attribute(fieldName, "nbins", compArr[1]["nbins"])

                elif compArr[0] == "huffman":  # differenced huffman
                    dictNum = compArr[1]["dictNum"]
                    if dictNum not in self.huffDict.keys():
                        self.huffDict[dictNum] = {}
                    delta = np.diff(data)
                    delta = np.insert(delta, 0, data[0])
                    self.huffDict[dictNum][fieldName] = delta
                    # print("adding " + fieldName + " to dict, contents ", delta[delta != 0], data[data != 0])
                    self.add_attribute(fieldName, "huffmanDictNumber", dictNum)
                    writeField = False

                elif compArr[0] == "rice":  # rice encoding
                    k = compArr[1]["k"]
                    data, k = rice.encode(data, k)
                    data = np.void(bytes(data))
                    self.add_encoding(fieldName.rsplit("/", 1)[0] + "/riceK", k)
                    self.add_attribute(fieldName, "riceK", k)

                else:
                    raise ValueError(
                        "Compression type "
                        + compArr[0]
                        + " is not a recognized compression"
                    )
            self.add_attribute(fieldName, "compression", compInfo)
            # print("adding " + compInfo + ' to ' + fieldName)

        if ":" in fieldName:
            writeField = False
            dictName, index = fieldName.rsplit(":", 1)
            if dictName not in self.raggedDict.keys():
                self.raggedDict[dictName] = {}

            self.raggedDict[dictName][index] = data

        if writeField:
            try:
                self.outFile.create_dataset(fieldName, data=data)
            except OSError as e:
                if self.overwrite:
                    del self.outFile[fieldName]
                    self.outFile.create_dataset(fieldName, data=data)
                else:
                    raise OSError(e)
            for attr in self.attrDict.copy().keys():
                fName, attrName = attr.split("@")
                if fName == fieldName:
                    self.add_attribute(fieldName, attrName, self.attrDict.pop(attr))

    def add_attribute(self, fieldName, attrName, data):
        try:
            self.outFile[fieldName].attrs[attrName] = data
        except KeyError as k:
            self.attrDict[fieldName + "@" + attrName] = data

    def add_encoding(self, encoding, value):
        if encoding not in self.encodings.keys():
            self.encodings[encoding] = value
        else:
            if self.encodings[encoding] != value:
                print(
                    "Warning: Inconsistant encoding value "
                    + encoding
                    + " is set to "
                    + str(self.encodings[encoding])
                    + " but wants to be "
                    + str(value)
                )

    def add_softlink(self, linkName, dataName):
        self.outFile[linkName] = h5py.SoftLink(dataName)

    def finalize_file(self):

        if not self.exists or self.overwrite:
            for encoding in self.encodings.keys():
                self.add_field(encoding, [self.encodings[encoding]])
                # print('adding ' + encoding + ' to file ' + self.outName)

            self.add_field("/common/version", self.version)
            # [Maksym]: was getting the error:
            # ...
            # File ".../python/commander_tools/tod_tools/commander_tod.py", line 213, in finaliz    e_file
            # self.add_field('/common/pids', list(self.pids.keys()))
            # File ".../python/commander_tools/tod_tools/commander_tod.py", line 179, in add_fie    ld
            # self.outFile.create_dataset(fieldName, data=data)
            # ...
            # File "h5py/h5t.pyx", line 1629, in h5py.h5t.py_create
            # File "h5py/h5t.pyx", line 1653, in h5py.h5t.py_create
            # File "h5py/h5t.pyx", line 1719, in h5py.h5t.py_create
            # TypeError: No conversion path for dtype: dtype('<U6')
            # So needed to add `np.string_()`
            self.add_field("/common/pids", np.string_(list(self.pids.keys())))

        if self.filelists is not None:
            for pid in self.pids.keys():
                self.filelists[self.freq]["id" + str(pid)] = (
                    str(pid)
                    + ' "'
                    + os.path.abspath(self.outName)
                    + '" '
                    + "1 "
                    + self.pids[pid]
                    + "\n"
                )

        return

    def finalize_chunk(self, pid, loadBalance=[0, 0]):
        if len(loadBalance) != 2:
            raise ValueError("Load Balancing numbers must be length 2")
        for key in self.huffDict.keys():
            h = huffman.Huffman()
            h.GenerateCode(list(self.huffDict[key].values()))
            huffArray = np.append(
                np.append(np.array(h.node_max), h.left_nodes), h.right_nodes
            )
            numStr = str(key)
            if key == 1:
                numStr = ""

            self.add_field(
                "/" + str(pid).zfill(6) + "/common/hufftree" + numStr, huffArray
            )
            self.add_field(
                "/" + str(pid).zfill(6) + "/common/huffsymb" + numStr, h.symbols
            )
            # with np.printoptions(threshold=np.inf):
            # print(huffArray, len(huffArray), len(h.symbols))
            for field in self.huffDict[key].keys():
                if ":" not in field:
                    self.add_field(
                        field, np.void(bytes(h.byteCode(self.huffDict[key][field])))
                    )
                    # print(self.huffDict[key][field], len(h.byteCode(self.huffDict[key][field])))
                else:
                    fieldName, index = field.rsplit(":", 1)
                    self.raggedDict[fieldName][index] = np.void(
                        bytes(h.byteCode(self.huffDict[key][field]))
                    )

        # handle the compressed ragged matrices
        for field in self.raggedDict.keys():
            ragged = False
            length = len(list(self.raggedDict[field].values())[0])
            for entry in self.raggedDict[field].keys():
                if len(self.raggedDict[field][entry]) != length or length < 1:
                    # np.void objects seem to have length 0
                    ragged = True

            if ragged == False:
                dataType = np.dtype(list(self.raggedDict[field].values())[0][0])
                array = np.ndarray(
                    (len(self.raggedDict[field].keys()), length), dtype=dataType
                )

                for entry in self.raggedDict[field].keys():
                    array[int(entry)] = self.raggedDict[field][entry]

                self.outFile.create_dataset(field, data=array)

            else:
                dt = h5py.special_dtype(vlen=np.dtype("uint8"))
                dset = self.outFile.create_dataset(
                    field, len(list(self.raggedDict[field].keys())), dtype=dt
                )
                for entry in self.raggedDict[field].keys():
                    dset[int(entry)] = np.frombuffer(
                        self.raggedDict[field][entry], dtype="uint8"
                    )

        for attr in self.attrDict.copy():
            field, attrName = attr.split("@", 1)
            if ":" in field:
                fieldName, index = field.split(":", 1)
            else:
                fieldName = field

            self.add_attribute(fieldName, attrName, self.attrDict.pop(attr))

        if len(self.attrDict) > 0:
            print("Attributes left unassigned at chunk end", self.attrDict)

        self.raggedDict = {}
        self.huffDict = {}
        self.add_field("/" + str(pid).zfill(6) + "/common/load", loadBalance)
        self.pids[pid] = str(float(loadBalance[0])) + " " + str(float(loadBalance[1]))

    def compute_version(self):
        return

    def make_filelists(self):
        for freq in self.filelists.keys():
            outfile = open(
                os.path.join(self.outPath, "filelist_" + str(freq) + ".txt"), "w"
            )
            outfile.write(str(len(self.filelists[freq])) + "\n")
            for buf in self.filelists[freq].values():
                # print(buf, len(buf))
                outfile.write(buf)

            outfile.close()

        return

    # File Reading Functions
    def load_field(self, fieldName):
        if fieldName[0] != "/":  # catch common user error
            fieldName = "/" + fieldName

        if 'compression' in self.outFile[fieldName].attrs.keys():
            compStr = self.outFile[fieldName].attrs["compression"]
            return self.decompress(fieldName, compression=compStr)
        else:
            return self.outFile[fieldName]


    def load_all_fields(self):
        return

    def read_across_files(self, fieldName):
        return

    def close(self):
        self.outFile.close()

    def decompress(self, field, compression=""):
        comps = compression.split(" ")
        ndim = 1

        matrix = False
        try:
            matrix = self.outFile[field].attrs["matrix"]
        except KeyError:
            ndim = 1

        if matrix:
            data = self.outFile[field][:]
            ndim = len(data)
        else:
            data = self.outFile[field]

        for i in range(ndim):

            if ndim > 1:
                dataBuf = data[i]
            else:
                dataBuf = data

            for comp in comps[::-1]:  # apply the filters in the reverse order
                if comp == "":
                    # residual from str.split()
                    pass
                elif comp == "dtype":
                    pass
                elif comp == "sigma":
                    sigma0 = self.outFile[field].attrs["sigma0"]
                    nsigma = self.outFile[field].attrs["nsigma"]
                    offset = self.outFile[field].attrs["offset"]
                    dataBuf = dataBuf * sigma0 / nsigma + offset

                elif comp == "digitize":
                    nbins = self.outFile[field].attrs["nbins"]
                    nmin = self.outFile[field].attrs["min"]
                    nmax = self.outFile[field].attrs["max"]

                    bins = np.linspace(nmin, nmax, num=nbins)
                    dataBuf = bins[dataBuf.astype('int')]

                elif comp == "huffman":
                    pid = field.split("/")[1]
                    try:
                        huffNum = str(self.outFile[field].attrs["huffmanDictNumber"])
                    except KeyError:
                        huffNum = ""
                    if huffNum == "1":
                        huffNum = ""
                    huffTree = self.load_field("/" + pid + "/common/hufftree" + huffNum)
                    huffSymb = self.load_field("/" + pid + "/common/huffsymb" + huffNum)
                    h = huffman.Huffman(tree=huffTree, symb=huffSymb)
                    dataBuf = h.Decoder(np.array(dataBuf))

                elif comp == "rice":
                    k = self.load_field(field.rsplit("/", 1)[0] + "/riceK")
                    dataBuf = np.array(rice.decode(np.array(dataBuf), k))
                    offset = dataBuf[0]
                    dataBuf += offset
                    dataBuf = dataBuf[1:]

                else:
                    raise ValueError(
                        "Decompression type " + comp + " is not a recognized operation"
                    )

            if ndim > 1:
                data[i] = dataBuf
            else:
                data = dataBuf

        return data



from __future__ import annotations

from abc import ABC
from dataclasses import dataclass
from typing import Protocol

import astropy.units as u
import numpy as np
from astropy.units import Quantity
from numpy.typing import NDArray


class FreqScalingFunc(Protocol):
    def __call__(
        self,
        freqs: Quantity[u.Hz] | Quantity[u.m],
        freq_ref: Quantity[u.Hz] | Quantity[u.m],
        **spectral_params: dict[str, float | Quantity | NDArray[np.float_]],
    ) -> float | NDArray[np.float_]:
        ...


@dataclass
class SkyComponent(ABC):
    amp_ref: Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]
    freq_ref: Quantity[u.Hz] | Quantity[u.m]
    spectral_params: dict[str, float | Quantity | NDArray[np.float_]]
    freq_scaling_func: FreqScalingFunc

    def get_delta_emission(
        self, freqs: Quantity[u.Hz] | Quantity[u.m]
    ) -> Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]:
        ...

    def get_bandpass_emission(
        self, freqs: Quantity[u.Hz] | Quantity[u.m], weights: NDArray[np.float_]
    ) -> Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]:
        ...


class Diffuse(SkyComponent):
    def get_delta_emission(
        self, freqs: Quantity[u.Hz] | Quantity[u.m]
    ) -> Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]:

        scale_factor = self.freq_scaling_func(
            freqs=freqs, 
            freq_ref=self.freq_ref,
            **self.spectral_params
        )
        return self.amp_ref * scale_factor

    def get_bandpass_emission(
        self, freqs: Quantity[u.Hz] | Quantity[u.m], weights: NDArray[np.float_]
    ) -> Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]:
        ...


@dataclass
class PointSource(SkyComponent):
    point_source_catalog: Quantity[u.deg]

    def get_delta_emission(
        self, freqs: Quantity[u.Hz] | Quantity[u.m]
    ) -> Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]:

        scale_factor = self.freq_scaling_func(
            freqs=freqs, 
            freq_ref=self.freq_ref,
            **self.spectral_params
        )

        scaled_source_points = self.amp_ref * scale_factor

        emission = bin_point_sources(
            
        )
        return self.amp_ref * scale_factor

    def get_bandpass_emission(
        self, freqs: Quantity[u.Hz] | Quantity[u.m], weights: NDArray[np.float_]
    ) -> Quantity[u.K] | Quantity[u.KR_J] | Quantity[u.Jy / u.sr]:
        ...


import click
from rich import print
import healpy as hp
import numpy as np

from cosmoglobe.h5 import chain
from cosmoglobe.h5.chain import Chain

@click.group()
def commands_h5():
    pass

# fmt: off
@commands_h5.command()
@click.argument("chain", type=click.Path(exists=True),)
@click.option( "-samples", default=[-1], multiple=True, help="",)
@click.option( "-new_name", type=click.Path(exists=True), default=None, help="Name of the chain copy. If None, a default is [chain_name]_copy.h5",)
@click.option( "-min", type=click.INT, help="Min sample number for range",)
@click.option( "-max", type=click.INT, help="Max sample number",)
# fmt: on
def copy_chain(**kwargs):
    """Creates a copy of the chain with a single or multiple samples.

    Parameters
    ----------
    chain
        Path to the chain file or the `Chain` object to copy.
    samples
        Samples to copy. Can be an int, a list of ints or a python range object.
    new_name
        Name of the chain copy. If None, a default is "{chain.name}_copy.h5"
    """
    if (kwargs['min'] is not None) or (kwargs['max'] is not None):
        kwargs['samples'] = range(kwargs['min'], kwargs['max'])
    elif kwargs['samples'] == [-1]:
        kwargs['samples']=int(kwargs['samples'][0])
    del kwargs['min']
    del kwargs['max']
    chain.copy_chain(**kwargs)

# fmt: off
@commands_h5.command()
@click.argument("chain", type=click.Path(exists=True),)
@click.argument("other_chain", type=click.Path(exists=True),)
@click.option( "-group_list", multiple=True, help="List of hdf5 groups that will be overwritten in the new chainfile. Use -group_list thing1 -group_list thing2.",)
@click.option( "-new_name", type=click.Path(exists=True), default=None, help="Name of the chain copy. If None, a default is [chain name]_copy.h5",)
# fmt: on
def combine_chains(**kwargs):
    """Creates a new chainfile that combines specific groups from two chains.

    The new file will contain all content from `chain`, except for the content
    within the groups in the `group_list`, which are taken from `other_chain`
    instead.

    Parameters
    ----------
    chain
        Path to chain file. This chain defines all the chain whos content you
        want to overwrite in a new combined file.
    other_chain
        Path to chain file.  This chain contains the groups you want to
        overwrite in `chain`.
    group_list
        List of hdf5 groups that will be overwritten in the new chainfile.
    new_name
        Name of the chain copy. If None, a default is "{chain.name}_copy.h5"
    """
    chain.combine_chains(**kwargs)

# fmt: off
@commands_h5.command()
@click.argument("chain", type=click.Path(exists=True),)
@click.argument("dataset",)
@click.argument("outname", )
@click.option( "-min", type=click.INT, default=0, help="First sample.",)
@click.option( "-max", type=click.INT, default=None,  help="Last sample. If not specified, use last sample of dataset.",)
@click.option( "-nside", type=click.INT, default=None,  help="Output nside of data.",)
# fmt: on
def mean(chain, dataset, outname, min, max, nside):
    """Calculates the mean of a dataset over a range of samples

    Parameters
    ----------
    chain
        Path to chain file. This chain defines all the chain whos content you
        want to overwrite in a new combined file.
    dataset
        Chain dataset on the form dust/amp_alm
    outname
        Name of .fits file to output the result to.
    min
        The first sample, default is first sample of chain.
    max
        last sample, default is last sample of chain.
    nside
        Output nside of data.
    """

    chain = Chain(chain)
    if max is None: max = chain.nsamples
    data = chain.mean(dataset, samples=range(min, max))

    if outname.endswith(".fits"):
        if dataset.endswith("alm"): 
            comp, quantity = dataset.split("/")
            if nside is None: nside = chain.parameters[comp]["nside"]
            pol = True if quantity.startswith("amp") else False
            fwhm = chain.parameters[comp]["fwhm"]
            data = hp.alm2map(data, nside=nside, fwhm=fwhm, pol=pol, pixwin=True,)
            
        hp.write_map(outname, data, overwrite=True)
    else:
        np.savetxt(outname, data)

# fmt: off
@commands_h5.command()
@click.argument("chain", type=click.Path(exists=True),)
@click.argument("dataset",)
@click.argument("outname", )
@click.option( "-min", type=click.INT, default=0, help="First sample.",)
@click.option( "-max", type=click.INT, default=None,  help="Last sample. If not specified, use last sample of dataset.",)
@click.option( "-nside", type=click.INT, default=None,  help="Output nside of data.",)
# fmt: on
def stddev(chain, dataset, outname, min, max, nside):
    """Calculates the stddev of a dataset over a range of samples

    Parameters
    ----------
    chain
        Path to chain file. This chain defines all the chain whos content you
        want to overwrite in a new combined file.
    dataset
        Chain dataset on the form dust/amp_alm
    outname
        Name of .fits file to output the result to.
    min
        The first sample, default is first sample of chain.
    max
        last sample, default is last sample of chain.
    nside
        Output nside of data.
    """

    chain = Chain(chain)
    if max is None: max = chain.nsamples

    if outname.endswith(".fits"):
        alm2map = True if dataset.endswith("alm") else False
        data = chain.stddev(dataset, samples=range(min, max), alm2map=alm2map)
        hp.write_map(outname, data, overwrite=True)
    else:
        data = chain.stddev(dataset, samples=range(min, max),)
        np.savetxt(outname, data)




import numpy as np
from astropy.units import Quantity

from cosmoglobe.data import DATA_DIR
from cosmoglobe.sky._base_components import PointSourceComponent
from cosmoglobe.sky.components._labels import SkyComponentLabel

RADIO_CATALOG = DATA_DIR / "radio_catalog.dat"


class AGNPowerLaw(PointSourceComponent):
    r"""Class representing the radio component.

    Notes
    -----
    This is a generic power law given at a reference frequency
    :math:`\nu_{\mathrm{0, src}}` with a power law spectral index
    :math:`\alpha`. It is defined using the convention in
    `BeyondPlanck (2020), Section 3.4.1
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{src}}(\nu) \propto
        \left(\frac{\nu}{\nu_{\mathrm{0, src}}}\right)^{\alpha-2}

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission.
    """

    label = SkyComponentLabel.RADIO
    catalog = Quantity(
        np.loadtxt(RADIO_CATALOG, usecols=(0, 1)).transpose(), unit="deg"
    )

    def get_freq_scaling(self, freqs: Quantity, alpha: Quantity) -> Quantity:
        """See base class."""

        scaling = (freqs / self.freq_ref) ** (alpha - 2)

        return scaling



#================================================================================
#
# Copyright (C) 2020 Institute of Theoretical Astrophysics, University of Oslo.
#
# This file is part of Commander3.
#
# Commander3 is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Commander3 is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Commander3. If not, see <https://www.gnu.org/licenses/>.
#
#================================================================================

import healpy as hp
import numpy as np
import heapq
import os
from time import time

_node_number = 0

class LeafNode:

    def __init__(self, symbol, weight):
        global _node_number
        _node_number += 1
        self.node_number = _node_number
        self.symbol = symbol
        self.weight = weight
        self.left = None
        self.right = None

    def __lt__(self, other):
        return self.weight < other.weight


class Huffman:

    def __init__(self, infile="", nside=256, tree=None, symb=None):
        self.infile = infile
        self.nside = nside
        self.encoding = {}
        self.decoding = {}
        self.queue = []
        self.weight = {}
        self.symbols = []
        self.left_nodes = []
        self.right_nodes = []
        self.node_max = 0
        self.head_node = None

        if(tree is not None):
            self.node_max = tree[0]
            self.left_nodes = tree[1:int((len(tree) -1)/2)+1]
            self.right_nodes = tree[int((len(tree) -1)/2)+1:]

        if(symb is not None):
            self.symbols = np.array(symb)

        if(symb is not None and tree is not None):
            self.BuildTree()
            self.PrintCode(self.head_node)

    #takes the input tree and symbols and generates the tree structure
    #input arrays are structured so that node numbers 1...nsymb correspond to 
    #the symbols array
    #the left and right arrays indicate the left and right children of nodes
    #nsymb+1 ... 2*nsymb-1
    def BuildTree(self):
        nodes = {}
        nsymb = len(self.symbols)

        self.head_node = LeafNode(None, 1)
        self.head_node.node_number = self.node_max
        self.head_node.left = self.left_nodes[self.node_max-nsymb-1]
        self.head_node.right = self.right_nodes[self.node_max-nsymb-1]
        nodes[self.node_max] = self.head_node

        for node in np.append(self.left_nodes, self.right_nodes):
            currNode = LeafNode(None, 1)
            currNode.node_number = node
            if node > nsymb:
                currNode.left = self.left_nodes[node -nsymb-1]
                currNode.right= self.right_nodes[node -nsymb-1]
            else:
                currNode.symbol = self.symbols[node-1]

            nodes[node] = currNode

        for node in nodes.copy().values():
            if node.left is not None:
                node.left = nodes[node.left]

            if node.right is not None:
                node.right = nodes[node.right]

    def PixellizePointing(self, diff=True, write=False):
        angs_pol = np.loadtxt(self.infile)
        pixels = hp.ang2pix(self.nside, angs_pol[:,1], angs_pol[:,0])

        if write :
            fname, fext = os.path.splitext(self.infile)
            np.save(fname+"_n"+str(self.nside).zfill(4)+"_pixels.npy", pixels)

        if diff:
            delta = np.diff(pixels)
            delta = np.insert(delta, 0, pixels[0])
            return delta
        else :
            return pixels

    def Weights(self, array):
        weight = {}
        for d in array:
            if not d in weight:
                    weight[d] = 0
            weight[d] += 1
        return weight

    def PrintCode(self, node, code=""):

        if(node.symbol != None):
            self.encoding[node.symbol] = code
            self.decoding[code] = node.symbol
            #print(node.symbol, node.weight, code)
            return
        self.PrintCode(node.left, code + "0")
        self.PrintCode(node.right, code + "1")


    def byteCode(self, array):
        text_bin = "".join(self.encoding[d] for d in array)
        padding = 8 - len(text_bin) % 8
        text_bin += padding*"0"
        # this tells you how many bits of padding there is apparently
        text_bin = "{0:08b}".format(padding) + text_bin

        b = bytearray()
        for i in range(0, len(text_bin), 8):
            byte = text_bin[i:i+8]
            b.append(int(byte, 2))
        return b


    def GenerateCode(self, array, write=False):
        array = np.array(array).flatten()
        self.weight = self.Weights(array)

        global _node_number

        for d in self.weight:
            node = LeafNode(d, self.weight[d])
            heapq.heappush(self.queue, node)
            self.symbols.append(d)
    
        while(len(self.queue)>1):
            left_child = heapq.heappop(self.queue)
            right_child = heapq.heappop(self.queue)

            collapsed = LeafNode(None, left_child.weight + right_child.weight)
            collapsed.left = left_child
            collapsed.right = right_child
            heapq.heappush(self.queue, collapsed)

            self.left_nodes.append(left_child.node_number)
            self.right_nodes.append(right_child.node_number)

        node = heapq.heappop(self.queue)
        if(node.left == None and node.right == None): #case where there was just one symbol
            self.PrintCode(node, code='0')
        else:
            self.PrintCode(node)

        self.node_max = node.node_number
        _node_number = 0

        b = self.byteCode(array)

        if write :
            fname, fext = os.path.splitext(self.infile)
            np.save(fname+"_n"+str(self.nside).zfill(4)+"_diffpix.npy", array)
            outfile = fname + "_n"+str(self.nside).zfill(4)+"_diffpix.bin"
            with open(outfile, 'wb') as f_out :
                f_out.write(bytes(b))
            return bytes(b), outfile
        else :
            return bytes(b)

    def Decoder(self, bytarr, write=False):
        bytarr = bytearray(bytarr)
        binary_txt = ''.join(bin(i)[2:].rjust(8,'0') for i in bytarr)
        padding = int(binary_txt[:8], 2)
        binary_txt = binary_txt[8:-1*padding]

        decoded_arr = []
        code = ""

        for b in binary_txt:
            code += b

            if code in self.decoding:
                decoded_arr.append(self.decoding[code])
                code = ""
        decoded_arr = np.cumsum(decoded_arr)

        if write:
            fname, fext = os.path.splitext(self.infile)
            file_out = fname + "_n"+str(self.nside).zfill(4)+"_decoded_pixels.npy"
            np.save(file_out, decoded_arr)
            return decoded_arr, file_out
        else:
            return decoded_arr



from dataclasses import dataclass, field
from typing import Dict, List, Type

from cosmoglobe.sky._base_components import SkyComponent
from cosmoglobe.sky._exceptions import (
    ComponentNotFoundError,
    CosmoglobeModelError,
    ModelNotFoundError,
)
from cosmoglobe.sky.components.ame import SpinningDust
from cosmoglobe.sky.components.cmb import CMB, CMB_LOWL
from cosmoglobe.sky.components.dust import ModifiedBlackbody
from cosmoglobe.sky.components.freefree import LinearOpticallyThin
from cosmoglobe.sky.components.radio import AGNPowerLaw
from cosmoglobe.sky.components.synchrotron import PowerLaw


@dataclass
class CosmoglobeModel:
    """The sky components making up Cosmoglobe Sky Model."""

    version: str
    components: List[Type[SkyComponent]]

    def __post_init__(self) -> None:
        """Makes sure that no duplicates of a component label exists."""

        if any(self.components.count(comp) > 1 for comp in self.components):
            raise CosmoglobeModelError(
                "components list to CosmoglobeModel cant contain multiple "
                "representations of the same labeled component."
            )

    def __getitem__(self, component_name: str) -> Type[SkyComponent]:
        """Returns a sky component from the cosmoglobe model."""

        for component in self.components:
            if component.label.value == component_name:
                return component
        raise ComponentNotFoundError(f"component {component_name} not found in model.")


@dataclass
class CosmoglobeModelRegistry:
    """Container for registered sky model versions."""

    REGISTRY: Dict[str, CosmoglobeModel] = field(default_factory=dict)

    def register_model(self, model: CosmoglobeModel) -> None:
        """Adds a new sky model to the registry."""

        if (version := model.version) in self.REGISTRY:
            raise ValueError(f"model by version {version} is already registered.")

        self.REGISTRY[version] = model

    def get_model(self, version: str) -> CosmoglobeModel:
        """Returns a registered sky model."""
        try:
            return self.REGISTRY[version]
        except KeyError:
            raise ModelNotFoundError(
                f"model {version} was not found in the registry. "
                f"Available models are: {list(self.REGISTRY.keys())}"
            )


cosmoglobe_registry = CosmoglobeModelRegistry()
cosmoglobe_registry.register_model(
    CosmoglobeModel(
        version="BeyondPlanck",
        components=[
            SpinningDust,
            CMB,
            CMB_LOWL,
            ModifiedBlackbody,
            LinearOpticallyThin,
            AGNPowerLaw,
            PowerLaw,
        ],
    )
)

DEFAULT_COSMOGLOBE_MODEL = cosmoglobe_registry.get_model("BeyondPlanck")




from re import A
import time
totaltime = time.time()
import click
import sys
import os
import healpy as hp
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as col
from cosmoglobe.release.tools import arcmin2rad
import os
# Fix for macos openMP duplicate bug
os.environ['KMP_DUPLICATE_LIB_OK']='True'

#print("Importtime:", (time.time() - totaltime))

def trygveplot(input, dataset=None, nside=None, auto=False, min=False, max=False, mid=[], rng="auto", colorbar=False,
            graticule=False, lmax=None, fwhm=0.0, mask=None, mfill=None, sig=[0,], remove_dipole=None, remove_monopole=None,
            logscale=False, size="m", white_background=False, darkmode=False, png=False, cmap=None, title=None,
            ltitle=None, unit=None, scale=None, outdir='.', outname=None, verbose=False, fontsize=11, gif=False,
            oldfont=False, dpi=300, xsize=2000, hires=False):
    """
    Plots a fits file, a h5 data structure or a data array in Mollview with pretty formatting.
    Option of autodecting component base on file-string with json look-up table.

    Parameters
    ----------
    input : array, str or list
        String or list of strings with filenames such as "cmb.fits" or "chain_c001.h5" for input data.
    dataset : str
        if input is a .h5 hdf-file specify which dataset to plot, for example "000007/cmb/amp_alm"
        default = None        
    nside : int
        nside for optional ud_grade (required when loading from hdf file)
        default = None
    auto : bool
        Toggle parameter autodetector.
        Automatic identification of plotting parameters based on information from autoparams.json
        default = False
    min :  str 
        Minimum value of colorbar. Overrides autodetector.
        default = False
    max :  str 
        Maximum value of colorbar. Overrides autodetector.
        default = False
    mid : list
        Adds tick values between min and max. Accepts multiple ex. -mid 2 -mid 3.
        default = []
    range : str
        Min and max value around 0 (ex. 10). 
        Also compatible with "auto": sets to 97.5 percentile of data.
        Also compatible with "minmax": sets to min and max values.
        default = "auto"
    colorbar : bool
        Adds a colorbar, and "cb" to output filename.
        default = False
    graticule : bool
        Adds graticule to figure.
        default = False
    lmax : float
        Set the lmax of alm input data from hdf.
        If reading hdf and not set, this is automatically found.
        default = None
    fwhm : float
        Optional map smoothing. FWHM of gaussian smoothing in arcmin.
        default = 0.0
    mask : str
        Apply a mask file (supply filename) to data (uses signal 0). 
        default = None
    mfill : str
        Color to fill masked area, for example "gray".
        Transparent by default.
        defualt : None
    sig : list of int
        Signal indices to be plotted. 0, 1, 2 interprated as IQU.
        Supports multiple, such that -sig 1 -sig 2.
        default = [0,]
    remove_dipole : str
        Fits and removes a dipole. Specify mask for fit, or "auto".
        default = None
    remove_monopole : str
        Fits and removes a monopole. Specify mask for fit, or "auto".
        default = None
    logscale : str
        Normalizes data using a semi-logscale linear between -1 and 1. 
        Autodetector uses this sometimes, you will be warned.
        default = None
    size : str
        Size of output maps. 
        1/3, 1/2 and full page width (8.8/12/18cm) [ex. x, s, m or l, or ex. slm for all]"
        default = "m"
    white_background : bool
        Sets white background. Transparent by default.
        default = False
    darkmode : bool
        Plots all outlines in white for dark backgrounds, and adds "dark" in filename.
        default = False
    png : bool
        Saves file as png as opposed to PDF. Overwritten by outname extension.
        default = False
    cmap : str
        Colormap (ex. sunburst, planck, jet). Both matplotliib and cmasher available as of now.
        Also supports qualitative plotly map, [ex. q-Plotly-4 (q for qualitative 4 for max color)]
        Sets planck as default.
        default = None
    title : str
        Sets the upper right title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    ltitle : str
        Sets the upper left title. Has LaTeX functionaliity (ex. $A_{s}$.)
        default = None
    unit : str
        Sets uniit under color bar. Has LaTeX functionaliity (ex. $\mu K$".
        default = None
    scale : float
        Scale data by multiplicative factor. [ex. 1e-6 for muK to K
        default = None
    outdir : path
        Optional different output directory.
        default = "."
    outname : str
        Output filename. If not specified, descriptive filename will be generated from chosen params.
        Extension determines filetype and overrides -png flag.
        default = None
    gif : bool
        Make gif is input is list of data.
        defaullt = False
    oldfont : bool
        Use DejaVu font as oppoosed to Times.
        default = False
    fontsize : int
        Fontsize
        default = 11
    dpi : int
        DPI (Dots per inch). For figure resolution.
        default = 300
    xsize : int
        Figure width in pixels. Uses xsize*(xsize/2)
        default = 2000
    hires : bool
        For high resolution figures. Sets dpi to 3000 and xsize to 10000.
        default = False
    verbose : bool
        Verbose mode. Mainly timer outputs.
        default = False
    """
    if hires:
        xsize=10000
        dpi=1000
    if not oldfont:
        plt.rcParams["mathtext.fontset"] = "stix"
    plt.rcParams["backend"] = "agg" if png else "pdf"
    if outname:
        plt.rcParams["backend"] = "agg" if outname.endswith("png") else "pdf"
    plt.rcParams["axes.linewidth"] = 1
    if darkmode:
        params = ["text.color", "axes.facecolor", "axes.edgecolor", "axes.labelcolor", "xtick.color",
                  "ytick.color", "grid.color", "legend.facecolor", "legend.edgecolor"]
        for p in params:
            plt.rcParams[p] = "white"

    ### Terminal output for 
    click.echo("")
    click.echo(click.style("{:#^48}".format(""), fg="green"))
    click.echo(click.style("Plotting",fg="green") + f" {input} dset={dataset}")

    ####   READ MAP   #####
    maps_, lmax, outfile, signal_labels = get_map(input, sig, dataset, nside, lmax, fwhm,)

    # Terminal output specified signals
    click.echo(click.style("Using signals ",fg="green") + f"{sig}")
    click.echo(click.style("{:#^48}".format(""), fg="green"))

    #### Main plotting loop ####
    imgs = [] # Store images for gif
    for i, maps in enumerate(maps_):
        if maps.ndim == 1: maps = maps.reshape(1,-1)
        for pl, polt, in enumerate(sig):

            #### Select data column  #####
            m = hp.ma(maps[pl])
            signal_label = signal_labels[polt] if signal_labels else get_signallabel(polt) 
            nsid = hp.get_nside(m)
            filename=input[i] if isinstance(input, list) else input
            #### Smooth  #####
            if float(fwhm) > 0:# and filename.endswith(".fits"):
                click.echo(click.style(f"Smoothing fits map to {fwhm} arcmin fwhm",fg="yellow"))
                m = hp.smoothing(m, fwhm=arcmin2rad(fwhm), lmax=lmax,)

            #### ud_grade #####
            if nside is not None:# and filename.endswith(".fits"):
                if nsid != nside:
                    click.echo(click.style(f"UDgrading map from {nsid} to {nside}", fg="yellow"))
                    m = hp.ud_grade(m, nside,)
            else:
                nside = nsid

            #### Remove monopole or dipole #####
            if remove_dipole or remove_monopole: m = remove_md(m, remove_dipole, remove_monopole, nside)

            #### Scaling factor #####
            if scale:
                datastring = f"{outfile}{outname}"
                if "chisq" in datastring:
                    click.echo(click.style(f"Scaling chisq data with dof={scale}",fg="yellow"))
                    m = (m-scale)/np.sqrt(2*scale)
                else:
                    click.echo(click.style(f"Scaling data by {scale}",fg="yellow"))
                    m *= scale

            #### Automatic variables #####
            if auto:
                (m, ttl, lttl, unt, ticks, cmp, lgscale,) = get_params(m, outfile, outname, signal_label,)
                symlg = False
                # Tick bug fix

                mn, md, mx= (ticks[0], None, ticks[-1])
                if not mid and len(ticks)>2:
                    if ticks[0]<ticks[1]  and ticks[-2]<ticks[-1]:
                        md = ticks[1:-1]
                    else:
                        ticks.pop(1)
            else:
                ttl = lttl = unt = ""
                mn = md = mx = None
                ticks = [False, False]
                lgscale = False
                symlg = False
                cmp = "planck"

            #if "diff" in outname+outfile:
            #    click.echo(click.style("Changing cmap to planck because diff map!", fg="yellow", blink=True, bold=True))
            #    cmp = "planck"

            #### Commandline priority ####
            if logscale != None: lgscale = logscale
            if title: ttl = title
            if ltitle: lttl = ltitle 
            if unit: unt = unit 

            #### Colorbar ticks ####
            ticks = get_ticks(m, ticks, mn, md, mx, min, mid, max, rng, auto)
            ticklabels = [fmt(i, 1) for i in ticks]
            
            # Terminal ouput
            click.echo(click.style("FWHM: ", fg="green") + f"{fwhm}")
            click.echo(click.style("nside: ", fg="green") + f"{nside}")
            click.echo(click.style("Ticks: ", fg="green") + f"{ticks}")
            click.echo(click.style("Unit: ", fg="green") + f"{unt}")
            click.echo(click.style("Title: ", fg="green") + f"{ttl}")
            
            #### Logscale ####
            if lgscale:
                if ticks[1] < 0:
                    symlg = True
                    m, ticks = apply_logscale(m, ticks, linthresh=1)
                else:
                    symlg = False
                    m, ticks = apply_logscale(m, ticks, linthresh=0)

            #### Color map #####
            cmap_ = get_cmap(cmap, cmp, logscale=lgscale)
            
            #### Projection ####
            grid_pix, longitude, latitude = project_map(nside, xsize=xsize, ysize=int(xsize/ 2.0),)

            #### Mask ##########
            grid_map, cmap_ = apply_mask(m, mask, grid_pix, mfill, polt, cmap_) if mask else (m[grid_pix], cmap_)


            for width in get_sizes(size):
                click.echo(click.style("Size: ", fg="green") + str(width))
                #### Make figure ####
                height = width / 2.0
                if colorbar: height *= 1.275 # Make sure text doesnt change with colorbar
                if gif:
                    # Hacky gif implementation
                    if i == 0:
                        fig = plt.figure(figsize=(cm2inch(width), cm2inch(height),),)
                        ax = fig.add_subplot(111, projection="mollweide")
                else:
                    fig = plt.figure(figsize=(cm2inch(width), cm2inch(height),),)
                    ax = fig.add_subplot(111, projection="mollweide")
                
                #norm=col.SymLogNorm(linthresh=1, base=10) if logscale else None
                image = plt.pcolormesh(longitude[::-1], latitude, grid_map, vmin=ticks[0], vmax=ticks[-1], rasterized=True, cmap=cmap_, shading='auto', animated=gif)

                # Save img for gif
                if gif: imgs.append([image])

                #### Graticule ####
                if graticule: apply_graticule(ax, width)
                ax.xaxis.set_ticklabels([]); ax.yaxis.set_ticklabels([]) # rm lonlat ticklabs
                
                #### Colorbar ####
                if colorbar: apply_colorbar(fig, image, ticks, ticklabels, unt,
                        fontsize, linthresh=1, logscale=lgscale,
                        symlogscale=symlg)
                
                #### Right Title ####
                plt.text(4.5, 1.1, r"%s" % ttl, ha="center", va="center", fontsize=fontsize,)
                
                #### Left Title ####
                plt.text(-4.5, 1.1, r"%s" % lttl, ha="center", va="center", fontsize=fontsize,)
                
                #### Output file ####
                plt.tight_layout()
                if gif: #output gif on last iteration only
                    if i==len(input)-1:
                        output_map(fig, outfile, png, fwhm, colorbar, mask, remove_dipole, darkmode, white_background,cmap_,nside,signal_label,width,outdir, gif, imgs, dpi, verbose, outname)
                else:
                    output_map(fig, outfile, png, fwhm, colorbar, mask, remove_dipole, darkmode, white_background,cmap_,nside,signal_label,width,outdir, gif, imgs, dpi, verbose, outname)
                    click.echo("Saved, closing fig")
                    plt.close()
                click.echo("Totaltime:", (time.time() - totaltime),) if verbose else None

def get_params(m, outfile, outname, signal_label,):
    outfile = os.path.split(outfile)[-1] # Remove path 
    outfile = f"{outfile}{outname}"
    datastring = f"{outfile}{signal_label}"
    sl = signal_label.split("_")[0]
    if sl in ["Q", "U", "QU"]:
        i = 1
    elif sl == "P":
        i = 2
    else:
        i = 0
    import json
    from pathlib import Path
    with open(Path(__file__).parent /'autoparams.json', 'r') as f:
        params = json.load(f)
    for label, comp in params.items():
        if tag_lookup(comp["tags"], datastring):
            if label=="freefree" and tag_lookup(["diff"], datastring): continue
            click.echo(click.style("{:-^48}".format(f"Detected {label} signal {sl} in {datastring}"),fg="yellow"))
            if label in ["residual", "freqmap",  "smap", "bpcorr"]:
                from re import findall 
                if label == "smap": tit = str(findall(r"tod_(.*?)_Smap", outfile)[0])
                if label == "freqmap": 
                    #TODO generalize this
                    if '023-WMAP_K' in outfile:
                        tit = r'\mathit K'
                    elif '030-WMAP_Ka' in outfile:
                        tit = r'\mathit{Ka}'
                    elif '040-WMAP_Q1' in outfile:
                        tit = r'\mathit Q1'
                    elif '040-WMAP_Q2' in outfile:
                        tit = r'\mathit Q2'
                    elif '060-WMAP_V1' in outfile:
                        tit = r'\mathit V1'
                    elif '060-WMAP_V2' in outfile:
                        tit = r'\mathit V2'
                    elif '090-WMAP_W1' in outfile:
                        tit = r'\mathit W1'
                    elif '090-WMAP_W2' in outfile:
                        tit = r'\mathit W2'
                    elif '090-WMAP_W3' in outfile:
                        tit = r'\mathit W3'
                    elif '090-WMAP_W4' in outfile:
                        tit = r'\mathit W4'
                    elif "30" in outfile:
                        tit = "30"
                    elif "44" in outfile:
                        tit = "44"
                    elif "70" in outfile:
                        tit = "70"
                    elif "353" in outfile:
                        tit = "353"
                if label == "bpcorr": 
                    tit = str(findall(r"tod_(.*?)_bpcorr", outfile)[0])
                if label == "residual":
                    if "WMAP" in outfile:
                        m   *= 1e3
                        tit = str(findall(r"WMAP_(.*?)_", outfile)[0])
                        comp["ticks"] = [[-6, 0, 6]]
                    elif "Haslam" in outfile:
                        tit = "Haslam"
                    else:
                        if "res_"in outfile:
                            tit = str(findall(r"res_(.*?)_", outfile)[0])
                        else:
                            tit = str(findall(r"residual_(.*?)_", outfile)[0])
                    if "_P_" in outfile and sl != "T":
                        comp["ticks"] = [[-10, 0, 10]]
                    if "857" in outfile:
                        m *= 2.2703e-6 
                        comp["ticks"] = [[-300,0,300]]
                    elif "Haslam" in outfile:
                        comp["ticks"] = [[-1e4, 0, 1e4]]
                comp["comp"] = r"${"+tit+"}$"
            comp["comp"] = comp["comp"].lstrip('0')    
            #print(i,sl,len(comp["cmap"]),len(comp["ticks"]))
            if i>=len(comp["cmap"]): i = len(comp["cmap"])-1
            comp["cmap"] = comp["cmap"][i]
            comp["ticks"] = comp["ticks"][i]
            ttl, lttl = get_title(comp,outfile,signal_label,)
            if comp["ticks"] == "auto": comp["ticks"] = get_percentile(m,97.5)
            if label == "chisq": ttl = r"$\chi^2$"
            if label == "bpcorr": ttl ="$s_{\mathrm{leak}}^{"+tit.lstrip('0')+"}}$"
            if comp["unit"]: comp["unit"] = r"$"+comp["unit"].replace('$','')+"$"
            return (m,  ttl, lttl, comp["unit"], comp["ticks"], comp["cmap"], comp["logscale"],)
    # If not recognized
    click.echo(click.style("{:-^48}".format(f"Map not recognized, plotting with min and max values"),fg="yellow"))
    comp = params["unidentified"]
    comp["comp"] = signal_label.split("_")[-1]
    ttl, lttl = get_title(comp,outfile,signal_label,)
    if comp["ticks"][i] == "auto": comp["ticks"] = get_percentile(m,97.5)
    comp["cmap"] = comp["cmap"][i]
    return (m,  ttl, lttl, comp["unit"], comp["ticks"], comp["cmap"], comp["logscale"],)

def symlog(m, linthresh=1.0):
    # Extra fact of 2 ln 10 makes symlog(m) = m in linear regime
    m = m/linthresh/(2*np.log(10))
    return np.log10(0.5 * (m + np.sqrt(4.0 + m * m)))

def apply_logscale(m, ticks, linthresh=1):
    click.echo(click.style("Applying semi-logscale", fg="yellow", blink=True, bold=True))
    if linthresh == 0:
        inds = (m > 0)
        m[inds] = np.log10(m[inds])
        m[~inds] = np.log10(ticks[0])
        new_ticks = np.log10(ticks)
    else:
        m = symlog(m,linthresh)
        new_ticks = []
        for i in ticks:
            new_ticks.append(symlog(i,linthresh))

    m = np.maximum(np.minimum(m, new_ticks[-1]), new_ticks[0])
    return m, new_ticks

def get_signallabel(x):
    if x == 0:
        return "I"
    if x == 1:
        return "Q"
    if x == 2:
        return "U"
    return str(x)

def get_sizes(size):
    sizes = []
    if "x" in size:
        sizes.append(7)
    if "s" in size:
        sizes.append(8.8)
    if "m" in size:
        sizes.append(12.0)
    if "l" in size:
        sizes.append(18.0)
    return sizes

def get_percentile(m, percentile):
    vmin = np.percentile(m, 100.0 - percentile)
    vmax = np.percentile(m, percentile)

    vmin = 0.0 if abs(vmin) < 1e-5 else vmin
    vmax = 0.0 if abs(vmax) < 1e-5 else vmax
    return [vmin, vmax]

def get_cmap(cmap, cmp, logscale=False):
    # Chose colormap manually
    if cmap == None:
        # If not defined autoset or goto planck
        cmap = cmp
    if "planck" in cmap:
        from pathlib import Path
        if "planck_log" in cmap: #logscale:
            # setup nonlinear colormap
            from matplotlib.colors import LinearSegmentedColormap
            class GlogColormap(LinearSegmentedColormap):
                name = cmap
                def __init__(self, cmap, vmin=-1e3, vmax=1e7):
                    self.cmap = cmap
                    self.N = self.cmap.N
                    self.vmin = vmin
                    self.vmax = vmax

                def is_gray(self):
                    return False

                def __call__(self, xi, alpha=1.0, **kw):
                    x = xi * (self.vmax - self.vmin) + self.vmin
                    yi = self.modsinh(x)
                    # range 0-1
                    yi = (yi + 3)/10.
                    return self.cmap(yi, alpha)

                def modsinh(self, x):
                    return np.log10(0.5*(x + np.sqrt(x**2 + 4)))

            cmap_path = Path(__file__).parent / "planck_cmap_logscale.dat"
            planck_cmap = np.loadtxt(cmap_path) / 255.0            
            cmap = col.ListedColormap(planck_cmap, "planck")
            cmap = GlogColormap(cmap)        
        else:
            cmap_path = Path(__file__).parent / "planck_cmap.dat"
            planck_cmap = np.loadtxt(cmap_path) / 255.0            
            if cmap.endswith("_r"):
                planck_cmap = planck_cmap[::-1]
            cmap = col.ListedColormap(planck_cmap, "planck")


    elif "wmap" in cmap:
        from pathlib import Path
        cmap_path = Path(__file__).parent / "wmap_cmap.dat"
        wmap_cmap = np.loadtxt(cmap_path) / 255.0            
        if cmap.endswith("_r"):
            planck_cmap = planck_cmap[::-1]
        cmap = col.ListedColormap(wmap_cmap, "wmap")
    elif cmap.startswith("q-"):
        import plotly.colors as pcol
        _, clab, *numvals = cmap.split("-")
        colors = getattr(pcol.qualitative, clab)
        if clab=="Plotly":
            #colors.insert(3,colors.pop(-1))
            colors.insert(0,colors.pop(-1))
            colors.insert(3,colors.pop(2))
        try:
            cmap = col.ListedColormap(colors[:int(numvals[0])], f'{clab}-{numvals[0]}')
            click.echo(click.style("Using qualitative colormap:", fg="yellow") + f" {clab} up to {numvals[0]}")
        except:
            cmap = col.ListedColormap(colors,clab)
            click.echo(click.style("Using qualitative colormap:", fg="yellow") + f" {clab}")
    elif cmap.startswith("black2"):
        cmap = col.LinearSegmentedColormap.from_list(cmap,cmap.split("2"))
    else:
        try:
            import cmasher
            cmap = eval(f"cmasher.{cmap}")
        except:
            try:
                from cmcrameri import cm
                cmap = eval(f"cm.{cmap}")
            except:
                cmap = plt.get_cmap(cmap)

    click.echo(click.style("Colormap:", fg="green") + f" {cmap.name}")
    return cmap

def get_title(comp, outfile, signal_label,):
    sl = signal_label.split("_")[0]
    datastring = f"{outfile}{signal_label}"
    if tag_lookup(["STDDEV","_stddev"],datastring):
        if comp["special"]:
            ttl = r"$\sigma_{\mathrm{" + comp["param"].replace("$","") + "}}$"
        else:
            ttl =   r"$\sigma_{\mathrm{" + comp["comp"] + "}}$"
        comp["cmap"] = "neutral"
        comp["ticks"] = "auto"
        comp["logscale"] = comp["special"] = False
    elif tag_lookup(["RMS","_rms",],datastring):
        ttl =  comp["param"] + r"$_{\mathrm{" + comp["comp"] + "}}^{\mathrm{RMS}}$" 
        comp["cmap"] = "neutral"
        comp["ticks"] = "auto"
        comp["logscale"] = comp["special"] = False
    elif tag_lookup(["mean"],datastring):
        ttl = r"$\langle $" + comp["param"] + r"$_{" + comp["comp"] + "}^{ }$" + r"$\rangle$"
    elif tag_lookup(["diff"],datastring):
        if "dx12" in outfile: difflabel = "\mathrm{2018}"
        difflabel = "\mathrm{NPIPE}" if "npipe" in outfile else ""
        ttl = r"$\Delta$ " + comp["param"] + r"$_{\mathrm{" + comp["comp"] + "}}^{" + difflabel + "}$"
    else:
        ttl =  comp["param"] + r"$_{\mathrm{" + comp["comp"] + "}}^{ }$" 
    try:
        ttl = comp["custom"]
    except:
        pass
    # Left signal label
    lttl = r"$" + sl +"$"
    if lttl == "$I$":
        lttl = "$T$"
    elif lttl == "$QU$":
        lttl= "$P$"

    ttl = r"$"+ttl.replace('$','')+"$"
    lttl = r"$"+lttl.replace('$','')+"$",
    if ttl == "$$": ttl =""
    if lttl == "$$": lttl =""
    return ttl, lttl

def get_map(input, sig, dataset, nside, lmax, fwhm,):
    # Get maps array if .h5 file
    signal_labels=None
    maps=[]
    input = [input] if isinstance(input, (str, np.ndarray)) else input
    if not input: print("No input specified"); sys.exit()
    for input_ in input:
        if isinstance(input_, np.ndarray):
            if not nside: nside=hp.get_nside(input_)
            if not lmax:  lmax = 2.5*nside
            outfile = "figure"
            if input_.ndim > 1:
                maps_ = input_[sig]
            else:
                maps_ = input_

        elif input_.endswith(".h5"):
            from cosmoglobe.release.commands_hdf import h5map2fits
            from cosmoglobe.release.tools import alm2fits_tool
            # Get maps from alm data in .h5
            if dataset.endswith("alm"):
                if not nside:
                    click.echo(click.style("Specify nside for h5 files",fg="red"))
                    sys.exit()

                click.echo(click.style("Converting alms to map",fg="green"))

                (maps_, _, _, _, outfile,) = alm2fits_tool(input_, dataset, nside, lmax, fwhm, save=False,)
                maps_=maps_[sig]    
            # Get maps from map data in .h5
            elif dataset.endswith("map"):
                click.echo(click.style("Reading map from hdf",fg="green"))
                (maps_, _, _, outfile,) = h5map2fits(input_, dataset, save=False)
                maps_=maps_[sig]

            # Found no data specified kind in .h5
            else:
                click.echo(click.style("Dataset not found. Breaking.",fg="red"))
                click.echo(click.style(f"Does {input_}/{dataset} exist?",fg="red"))
                sys.exit()

        elif input_.endswith(".fits"):
            maps_, header = hp.read_map(input_, field=sig, h=True, dtype=None,)
            header = dict(header)
            signal_labels = []
            for i in range(int(header["TFIELDS"])):
                signal_label = header[f"TTYPE{i+1}"]
                if signal_label in ["TEMPERATURE", "TEMP"]:
                    signal_label = "T"
                if signal_label in ["Q-POLARISATION", "Q_POLARISATION"]:
                    signal_label = "Q"
                if signal_label in ["U-POLARISATION", "U_POLARISATION"]:
                    signal_label = "U"
                signal_labels.append(signal_label)            
            outfile = input_.replace(".fits", "")
        else:
            click.echo(click.style("Did not recognize data.",fg="red"))
            sys.exit()

        if maps_.ndim == 1: maps_ = maps_.reshape(1,-1)
        maps.append(maps_)
    return maps, lmax, outfile, signal_labels

def get_ticks(m, ticks, mn, md, mx, min, mid, max, rng, auto):
    # If min and max have been specified, set.
    if rng == "auto" and not auto: # Mathew: I have changed this to give the desired behaviour for make_diff_plots, but I don't know if this breaks other functionality
    # it used to be
    # if rng == "auto" and not auto:
        click.echo(click.style("Setting range from 97.5th percentile of data",fg="yellow"))
        mn,mx = get_percentile(m, 97.5)
    elif rng == "minmax":
        click.echo(click.style("Setting range from min to max of data",fg="yellow"))
        mn = np.min(m)
        mx = np.max(m)
    else:
        try:
            if float(rng)>0.0:
                mn = -float(rng)
                mx = float(rng)
                ticks = [False, 0.0, False]
        except:
            pass

    if min is False:
        min = mn
    else:
        min = float(min)

    if max is False:
        max = mx
    else:
        max = float(max)   
    ticks[0] = min
    ticks[-1] = max
    if mid:
        ticks = [min, *mid, max]
    elif md:
        ticks = [min, *md, max] 

    return [float(i) for i in ticks]

def apply_mask(m, mask, grid_pix, mfill, polt, cmap):
    click.echo(click.style(f"Masking using {mask}", fg="yellow"))
    # Apply mask
    hp.ma(m)
    mask_field = polt-3 if polt>2 else polt
    m.mask = np.logical_not(hp.read_map(mask, field=mask_field, dtype=None))

    # Don't know what this does, from paperplots by Zonca.
    grid_mask = m.mask[grid_pix]
    grid_map = np.ma.MaskedArray(m[grid_pix], grid_mask)

    if mfill:
        cmap.set_bad(mfill)  # color of missing pixels
        # cmap.set_under("white") # color of background, necessary if you want to use
        # using directly matplotlib instead of mollview has higher quality output

    return grid_map, cmap

def apply_colorbar(fig, image, ticks, ticklabels, unit, fontsize, linthresh,
        logscale, symlogscale):
    click.echo(click.style("Applying colorbar", fg="yellow"))
    from matplotlib.ticker import FuncFormatter
    cb = fig.colorbar(image, orientation="horizontal", shrink=0.4, pad=0.04, ticks=ticks, format=FuncFormatter(fmt),)
    cb.ax.set_xticklabels(ticklabels)
    cb.ax.xaxis.set_label_text(unit)
    cb.ax.xaxis.label.set_size(fontsize)



    if symlogscale:
        linticks = np.linspace(-1, 1, 3)*linthresh
        logmin = np.round(ticks[0])
        logmax = np.round(ticks[-1])

        logticks_min = -10**np.arange(np.log10(linthresh), abs(logmin)+1)
        logticks_max = 10**np.arange(np.log10(linthresh), logmax+1)
        ticks_ = np.unique(np.concatenate((logticks_min, linticks, logticks_max)))
        #cb.set_ticks(np.concatenate((ticks,symlog(ticks_))), []) # Set major ticks

        logticks = symlog(ticks_, linthresh)
        logticks = [x for x in logticks if x not in ticks]
        cb.set_ticks(np.concatenate((ticks,logticks ))) # Set major ticks
        cb.ax.set_xticklabels(ticklabels + ['']*len(logticks))

        minorticks = np.linspace(-linthresh, linthresh, 5)
        minorticks2 = np.arange(2,10)*linthresh

        for i in range(len(logticks_min)):
            minorticks = np.concatenate((-10**i*minorticks2,minorticks))
        for i in range(len(logticks_max)):
            minorticks = np.concatenate((minorticks, 10**i*minorticks2))

        minorticks = symlog(minorticks, linthresh)
        minorticks = minorticks[ (minorticks >= ticks[0]) & ( minorticks<= ticks[-1]) ] 
        cb.ax.xaxis.set_ticks(minorticks, minor=True)
    elif logscale:
        logmin = np.round(ticks[0])
        logmax = np.round(ticks[-1])

        logticks_max = 10**np.arange(logmin, logmax)
        logticks = np.arange(logmin-1, logmax+2)
        minorticks = np.array([])
        minorticks2 = np.arange(1,11)
        for i in logticks:
            minorticks = np.concatenate((minorticks, 10**i*minorticks2))
        minorticks = np.log10(minorticks)
        minorticks = minorticks[(minorticks > ticks[0]) & (minorticks < ticks[-1])]
        cb.ax.xaxis.set_ticks(minorticks, minor=True)
        newlabs = []
        ticklabs_arr = np.array([t[1:-1] for t in ticklabels]).astype(float)
        for i in range(len(ticklabs_arr)):
            newlabs.append(np.log10(ticklabs_arr[i]))
        cb.ax.xaxis.set_ticks(np.array(newlabs))
        cb.ax.set_xticklabels(ticklabels)
    cb.ax.tick_params(which="both", axis="x", direction="in", labelsize=fontsize-2,)
    cb.ax.xaxis.labelpad = 0
    # workaround for issue with viewers, see colorbar docstring
    cb.solids.set_edgecolor("face")
    return cb

def apply_graticule(ax, width):
    click.echo(click.style("Applying graticule", fg="yellow"))
    #### Formatting ######    
    from matplotlib.projections.geo import GeoAxes
    class ThetaFormatterShiftPi(GeoAxes.ThetaFormatter):
        #shifts labelling by pi
        #Shifts labelling from -180,180 to 0-360
        def __call__(self, x, pos=None):
            if x != 0:
                x *= -1
            if x < 0:
                x += 2 * np.pi
            return GeoAxes.ThetaFormatter.__call__(self, x, pos)

    ax.set_longitude_grid(60)
    ax.xaxis.set_major_formatter(ThetaFormatterShiftPi(60))
    if width < 10:
        ax.set_latitude_grid(45)
        ax.set_longitude_grid_ends(90)            
        ax.grid(True)
        
def output_map(fig, outfile, png, fwhm, colorbar, mask, remove_dipole, darkmode, white_background,cmap,nside,signal_label,width,outdir,gif,imgs,dpi, verbose, outname):
    #### filename ##
    outfile = outfile.replace("_IQU_", "_")
    outfile = outfile.replace("_I_", "_")
    filename = []
    filename.append(f"{str(int(fwhm))}arcmin") if float(fwhm) > 0 else None
    filename.append("cb") if colorbar else None
    filename.append("masked") if mask else None
    filename.append("nodip") if remove_dipole else None
    filename.append("dark") if darkmode else None
    filename.append(f"c-{cmap.name}")

    nside_tag = "_n" + str(int(nside))
    if nside_tag in outfile:
        outfile = outfile.replace(nside_tag, "")

    fn = outfile + f"_{signal_label}_w{str(int(width))}" + nside_tag

    for i in filename:
        fn += f"_{i}"


    filetype = "png" if png else "pdf"
    if outname:
        filetype = "png" if outname.endswith(".png") else "pdf"

    fn += f".{filetype}"
    if outname: fn = outname
    starttime = time.time()
    if outdir:
        fn = outdir + "/" + os.path.split(fn)[-1]

    tp = False if white_background else True  
    if gif:
        click.echo(click.style("Outputting GIF:", fg="green") + f" {fn}")
        import matplotlib.animation as animation
        interval = 500 #100
        ani = animation.ArtistAnimation(fig, imgs, interval=interval, blit=True, repeat_delay=1000)
        ani.save(fn.replace(filetype,"gif"),dpi=dpi)
    else:
        click.echo(click.style("Outputting file", fg="green") + f" {fn}")
        fig.savefig(fn, bbox_inches="tight", pad_inches=0.02, transparent=tp, format=filetype, dpi=dpi)
    click.echo("Savefig", (time.time() - starttime),) if verbose else None

def fmt(x, pos):
    """
    Format color bar labels
    """
    if abs(x) > 1e4 or (abs(x) <= 1e-3 and abs(x) > 0):
        a, b = f"{x:.2e}".split("e")
        b = int(b)
        if float(a) == 1.00:
            return r"$10^{"+str(b)+"}$"
        elif float(a) == -1.00:
            return r"$-10^{"+str(b)+"}$"
        else:
            return fr"${a} \cdot 10^{b}$"
    elif abs(x) > 1e1 or (float(abs(x))).is_integer():
        return fr"${int(x):d}$"
    else:
        x = round(x, 2)
        return fr"${x}$"

def cm2inch(cm):
    return cm * 0.393701

def tag_lookup(tags, outfile):
    return any(e.lower() in outfile.lower() for e in tags)

def project_map(nside, xsize, ysize,):
    theta = np.linspace(np.pi, 0, ysize)
    phi = np.linspace(-np.pi, np.pi, xsize)
    longitude = np.radians(np.linspace(-180, 180, xsize))
    latitude = np.radians(np.linspace(-90, 90, ysize))
    # project the map to a rectangular matrix xsize x ysize
    PHI, THETA = np.meshgrid(phi, theta)
    grid_pix = hp.ang2pix(nside, THETA, PHI)

    return grid_pix, longitude, latitude

def remove_md(m, remove_dipole, remove_monopole, nside):        
    if remove_monopole:
        dip_mask_name = remove_monopole
    if remove_dipole:
        dip_mask_name = remove_dipole
    # Mask map for dipole estimation
    if dip_mask_name == 'auto':
        mono, dip = hp.fit_dipole(m, gal_cut=30)
    else:
        m_masked = hp.ma(m)
        m_masked.mask = np.logical_not(hp.read_map(dip_mask_name,dtype=None,))

        # Fit dipole to masked map
        mono, dip = hp.fit_dipole(m_masked)

    # Subtract dipole map from data
    if remove_dipole:
        click.echo(click.style("Removing dipole:", fg="yellow"))
        click.echo(click.style("Dipole vector:",fg="green") + f" {dip}")
        click.echo(click.style("Dipole amplitude:",fg="green") + f" {np.sqrt(np.sum(dip ** 2))}")

        # Create dipole template
        nside = int(nside)
        ray = range(hp.nside2npix(nside))
        vecs = hp.pix2vec(nside, ray)
        dipole = np.dot(dip, vecs)
        
        m = m - dipole
    if remove_monopole:
        click.echo(click.style("Removing monopole:", fg="yellow"))
        click.echo(click.style("Mono:",fg="green") + f" {mono}")
        m = m - mono
    return m

"""

def apply_colorbar(fig, image, ticks, unit, fontsize, norm):
    click.echo(click.style("Applying colorbar", fg="yellow"))
    from matplotlib.ticker import FuncFormatter
    cb = fig.colorbar(image, orientation="horizontal", shrink=0.4, pad=0.04, format=FuncFormatter(fmt),)
    ticklabels =  [fmt(i,1) for i in ticks]
    print(cb.ax.get_xticklabels())
    click.echo(click.style("Applying colorbar", fg="yellow"))
    from matplotlib.ticker import FuncFormatter, LogLocator
    cb = fig.colorbar(image, orientation="horizontal", shrink=0.4, pad=0.04, ticks=ticks, format=FuncFormatter(fmt),)

    cb.ax.set_xticklabels(ticklabels)
    cb.ax.xaxis.set_label_text(unit)
    cb.ax.xaxis.label.set_size(fontsize)
    if logscale:
        linticks = np.linspace(-1, 1, 3)*linthresh
        logmin = np.round(ticks[0])
        logmax = np.round(ticks[-1])

        logticks_min = -10**np.arange(0, abs(logmin)+1)
        logticks_max = 10**np.arange(0, logmax+1)
        ticks_ = np.unique(np.concatenate((logticks_min, linticks, logticks_max)))
        #cb.set_ticks(np.concatenate((ticks,symlog(ticks_))), []) # Set major ticks

        logticks = symlog(ticks_, linthresh)
        logticks = [x for x in logticks if x not in ticks]
        cb.set_ticks(np.concatenate((ticks,logticks ))) # Set major ticks
        cb.ax.set_xticklabels(ticklabels + ['']*len(logticks))

        minorticks = np.linspace(-linthresh, linthresh, 5)
        minorticks2 = np.arange(2,10)*linthresh

        for i in range(len(logticks_min)):
            minorticks = np.concatenate((-10**i*minorticks2,minorticks))
        for i in range(len(logticks_max)):
            minorticks = np.concatenate((minorticks, 10**i*minorticks2))

        minorticks = symlog(minorticks, linthresh)
        minorticks = minorticks[ (minorticks >= ticks[0]) & ( minorticks<= ticks[-1]) ] 
        cb.ax.xaxis.set_ticks(minorticks, minor=True)

    cb.ax.tick_params(which="both", axis="x", direction="in", labelsize=fontsize,)
    cb.ax.xaxis.labelpad = 0
    # workaround for issue with viewers, see colorbar docstring
    cb.solids.set_edgecolor("face")
    return cb
"""



from astropy.units import Quantity, Unit

from cosmoglobe.sky._base_components import DiffuseComponent
from cosmoglobe.sky.components._labels import SkyComponentLabel


class CMB(DiffuseComponent):
    r"""Class representing the CMB component in the Cosmoglobe Sky Model.

    Notes
    -----
    The CMB emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.2
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{CMB}}(\nu) \propto
        \frac{x^{2} \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)
        ^{2}} \boldsymbol{s}^{\mathrm{CMB}},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`x=h \nu / k T_{0}` and
    :math:`T_0 = 2.7255 \mathrm{K}` as of BP9.
    """

    label = SkyComponentLabel.CMB
    freq_range = (0 * Unit("Hz"), 1 * Unit("THz"))

    def get_freq_scaling(self, *_) -> Quantity:
        """See base class.

        NOTE: The CMB amplitude maps are stored in units of uK_CMB for which
        they are constant over frequencies.
        """

        return Quantity(1)

class CMB_LOWL(DiffuseComponent):
    r"""Class representing the CMB component in the Cosmoglobe Sky Model.

    Notes
    -----
    The CMB emission is defined using the convention in
    `BeyondPlanck (2020), Section 3.2
    <https://arxiv.org/pdf/2011.05609.pdf>`_;

    .. math::

        \boldsymbol{s}_{\mathrm{RJ}}^{\mathrm{CMB}}(\nu) \propto
        \frac{x^{2} \mathrm{e}^{x}}{\left(\mathrm{e}^{x}-1\right)
        ^{2}} \boldsymbol{s}^{\mathrm{CMB}},

    where :math:`\nu` is the frequency for which we are simulating the
    sky emission, :math:`x=h \nu / k T_{0}` and
    :math:`T_0 = 2.7255 \mathrm{K}` as of BP9.
    """

    label = SkyComponentLabel.CMB_LOWL
    freq_range = (0 * Unit("Hz"), 1 * Unit("THz"))

    def get_freq_scaling(self, *_) -> Quantity:
        """See base class.

        NOTE: The CMB amplitude maps are stored in units of uK_CMB for which
        they are constant over frequencies.
        """

        return Quantity(1)
